"""
alignPrincipalAxes.py

Generates rotation-translation matrices (Chimera format) for densities in mrc format
which if loaded in Chimera superimpose the density principal axes of inertia with
the x, y and z axes.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import mrcfile
import numpy as np
import os
import json
import argparse


def calc_principal_axis(xyz):
    """
    Calculates the principal axes of inertia of a density. Intensities are
    treated as masses. Theoretical basis: "Classical Mechanics" by J. B. Tatum
    :param xyz: density intensities as a numpy array
    :return: Coordinates of the center of the mass (Mc), principal moments of
    inertia (e_values), principal axes of inertia (e_vectors).
    """
    # Intensities lower than 0.2*maximum are set to 0 to reduce the noise and to
    # exclude negative values.
    max_int = np.amax(xyz)
    it = np.nditer(xyz, flags=['multi_index'], op_flags=['writeonly'])
    while not it.finished:
        if it[0] < max_int*0.2:
            it[0] = 0
        it.iternext()
    # Find coordinates of the centre of the mass, moments and products of inertia
    mass = np.sum(xyz)
    summ_xc = 0
    summ_yc = 0
    summ_zc = 0
    A = 0
    B = 0
    C = 0
    F = 0
    G = 0
    H = 0
    it = np.nditer(xyz, flags=['multi_index'], op_flags=['writeonly'])
    while not it.finished:
        summ_xc += it.multi_index[2] * it[0]
        summ_yc += it.multi_index[1] * it[0]
        summ_zc += it.multi_index[0] * it[0]
        # Find the moments and products of inertia relative to origin (0,0,0)
        A += (it.multi_index[1] ** 2 + it.multi_index[0] ** 2) * it[0]
        B += (it.multi_index[2] ** 2 + it.multi_index[0] ** 2) * it[0]
        C += (it.multi_index[2] ** 2 + it.multi_index[1] ** 2) * it[0]
        F += it.multi_index[1] * it.multi_index[0] * it[0]
        G += it.multi_index[2] * it.multi_index[0] * it[0]
        H += it.multi_index[2] * it.multi_index[1] * it[0]
        it.iternext()
    xc = summ_xc / mass
    yc = summ_yc / mass
    zc = summ_zc / mass
    # Coordinates of the center of the mass
    Mc = np.array([xc, yc, zc])
    # Find the moments and products of inertia relative to origin the center of the mass
    A = A - mass * (yc**2 + zc**2)
    B = B - mass * (xc**2 + zc**2)
    C = C - mass * (xc**2 + yc**2)
    # Products of inertia
    F = F - mass * yc * zc
    G = G - mass * xc * zc
    H = H - mass * xc * yc
    # Find principal moments and axes of inertia
    I = np.array([[A, -H, -G], [-H, B, -F], [-G, -F, C]])
    e_values, e_vectors = np.linalg.eig(I)
    # Sort axes based on principal moments values
    idx = e_values.argsort()[::1]
    e_values = e_values[idx]
    e_vectors = e_vectors[:, idx]
    # Make sure that a right handed set was obtained
    det_e_v = np.linalg.det(e_vectors)
    if det_e_v < 0:
        for i in range(3):
            e_vectors[i,0] = -e_vectors[i,0]
    return Mc, e_values, e_vectors


def calc_rot_trans(Mc, e_vectors, voxel_size, flip=None):
    """
    Calculates rotation-translation matrices in Chimera format. The principal axes
    of inertia are aligned to x, y and z axes of the original box. The center of
    the mass is aligned to the origin of the coordinate system. Since the directions
    of the axes ere arbitrary chosen the axes can be flipped.
    :param Mc: Coordinates of the center of the mass
    :param e_vectors: sorted principal axes of inertia
    :param voxel_size: density voxel size
    :param flip: axes to be flipped (None, x, y or xy)
    :return: rotation-translation matrix
    """
    flip_x = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
    flip_y = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, -1]])
    if flip is None:
        rot = np.transpose(e_vectors)
        trans = ([0,0,0] - np.dot(Mc, e_vectors)) * voxel_size
        rot_trans = np.hstack((rot, trans.reshape((3,1))))
    elif flip == "x":
        e_vectors = np.dot(e_vectors, flip_x)
        rot = np.transpose(e_vectors)
        trans = ([0, 0, 0] - np.dot(Mc, e_vectors)) * voxel_size
        rot_trans = np.hstack((rot, trans.reshape((3, 1))))
    elif flip == "y":
        e_vectors = np.dot(e_vectors, flip_y)
        rot = np.transpose(e_vectors)
        trans = ([0, 0, 0] - np.dot(Mc, e_vectors)) * voxel_size
        rot_trans = np.hstack((rot, trans.reshape((3, 1))))
    elif flip == "xy":
        e_vectors = np.dot(e_vectors, flip_x)
        e_vectors = np.dot(e_vectors, flip_y)
        rot = np.transpose(e_vectors)
        trans = ([0, 0, 0] - np.dot(Mc, e_vectors)) * voxel_size
        rot_trans = np.hstack((rot, trans.reshape((3, 1))))
    else:
        raise ValueError('Please input "x", "y", or "xy" or skip the flip parameter.')

    rot_trans = np.around(rot_trans, decimals=6)
    return rot_trans


def main():
    parser = argparse.ArgumentParser(description='Generate rotation-translation matrices for '
                                                 'principal component density alignment.')
    parser.add_argument("-f", "--file", dest="mrc_path", required=True,
                        help="Path a directory with densities in mrc format")
    args = parser.parse_args()
    work_dir = args.mrc_path

    files = os.listdir(work_dir)
    mrc_files = [i for i in files if i.endswith('.mrc')]
    length = len(mrc_files)
    moment_list = []
    norm_moment_list = []
    for k, element in enumerate(mrc_files):
        print('Working on', k + 1, 'file out of', length + 1, 'files.      "', element, '"')
        map1 = os.path.join(work_dir, element)
        with mrcfile.open(map1, mode='r+', permissive=True) as mrc:
            xyz = mrc.data
            voxel_size1 = mrc.header.cella.x / mrc.header.nx
        Mc1, e_val1, e_vec1 = calc_principal_axis(xyz)
        e_val1_norm = e_val1 / np.linalg.norm(e_val1)
        moment_list.append([element] + e_val1.tolist())
        norm_moment_list.append([element] + e_val1_norm.tolist())

        rot_trans1 = calc_rot_trans(Mc1, e_vec1, voxel_size1)
        rot_trans2 = calc_rot_trans(Mc1, e_vec1, voxel_size1, 'x')
        rot_trans3 = calc_rot_trans(Mc1, e_vec1, voxel_size1, 'y')
        rot_trans4 = calc_rot_trans(Mc1, e_vec1, voxel_size1, 'xy')
        matrix_list = [rot_trans1.tolist(), rot_trans2.tolist(), rot_trans3.tolist(), rot_trans4.tolist()]
        j_file = os.path.join(work_dir, element.split('.')[0] + '.json')
        # Save matrices as json files
        with open(j_file, 'w')as j:
            json.dump(matrix_list, j, indent=4)
        matrix_file = os.path.join(work_dir, element.split('.')[0] + '.txt')
        # Save matrices as text file loadable by Chimera
        with open(matrix_file, 'w') as fo:
            for i, matrix in enumerate(matrix_list):
                fo.write('Model ' + str(i) + '.0\n')
                for line in matrix:
                    fo.write(str(line[0]) + '  ' + str(line[1]) + '  ' + str(line[2]) + '  ' + str(line[3]) + '\n')

    moment_file = os.path.join(work_dir, 'a_principal_moments.txt')
    with open(moment_file, 'w') as m_f:
        for line in moment_list:
            text = ''
            for k in range(4):
                if k > 0:
                    line[k] = '{:.6f}'.format(round(line[k], 6))
                    text = text + '{0:>{width}}'.format(line[k], width=18)
                else:
                    text = text + '{0:>{width}}'.format(line[k], width=25)
            m_f.write(text + '\n')

    norm_moment_file = os.path.join(work_dir, 'a_principal_moments_norm.txt')
    with open(norm_moment_file, 'w') as m_f:
        for line in norm_moment_list:
            text = ''
            for k in range(4):
                if k > 0:
                    line[k] = '{:.6f}'.format(round(line[k], 6))
                    text = text + '{0:>{width}}'.format(line[k], width=18)
                else:
                    text = text + '{0:>{width}}'.format(line[k], width=25)
            m_f.write(text + '\n')


if __name__ == '__main__':
    main()


from chimerax.core.commands import Command
from shutil import copy2
import json
import os

rc = Command(session)

align_atoms_dict = {
    'symmetric': {'HIS': [['n', 'c', 'ca', 'cb', 'cg', 'nd1', 'ce1', 'cd2', 'ne2'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ne2', 'nd1', 'ce1']],
                  'PHE': [['n', 'c', 'ca', 'cb', 'cg', 'cd1', 'ce1', 'cd2', 'ce2', 'cz'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ce2', 'cd1', 'ce1', 'cz']],
                  'TYR': [['n', 'c', 'ca', 'cb', 'cg', 'cd1', 'ce1', 'cd2', 'ce2', 'cz'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ce2', 'cd1', 'ce1', 'cz']]
                  },
    'asymmetric': {'ASP': ['n', 'c', 'ca', 'cb', 'cg'],
                   'GLU': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'ASN': ['n', 'c', 'ca', 'cb', 'cg'],
                   'ARG': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'LYS': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'MET': ['n', 'c', 'ca', 'cb', 'cg', 'sd'],
                   'GLN': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'ILE': ['n', 'c', 'ca', 'cb', 'cg1', 'cd1'],
                   'LEU': ['n', 'c', 'ca', 'cb', 'cg', 'cd1'],
                   'TRP': ['n', 'c', 'ca', 'cb', 'cg', 'cd1'],
                   'PRO': ['n', 'c', 'ca', 'cb', 'cg'],
                   'THR': ['n', 'c', 'ca', 'cb', 'cg2'],
                   'VAL': ['n', 'c', 'ca', 'cb', 'cg1', 'cg2'],
                   'SER': ['n', 'c', 'ca', 'cb', 'og'],
                   'CYS': ['n', 'c', 'ca', 'cb', 'sg'],
                   'GLY': ['n', 'c', 'ca'],
                   'ALA': ['n', 'c', 'ca', 'cb']
                   }
                    }

with open('class_dict.json') as j_file:
    class_dict = json.load(j_file)

global_ref_path = class_dict['global_reference']
grid = class_dict['grid_path']
fit_map = class_dict['fit_map']
inp_dir = class_dict['inp_dir']
out_dir = class_dict['out_dir']
reference_map_path = class_dict['reference_mrc']
reference_model_path = class_dict['reference_model']

# Align the class reference residue and map to the global reference residue
rc.run('open ' + global_ref_path)
rc.run('open ' + reference_map_path)
model = rc.run('open ' + reference_model_path)[0][0]
rc.run('open ' + grid)
rc.run('align #3@ca,n,c to #1@ca,n,c')
rc.run('view position #2 sameAsModels #3')
rc.run('vop  resample #2 onGrid #4')

reference_map_path = os.path.join(out_dir, 'new_ref.mrc')
reference_model_path = os.path.join(out_dir, 'new_ref.cif')
rc.run(f'save {reference_map_path} models #5')
rc.run(f'save {reference_model_path} format mmcif models #3')

res_type = model.residues[0].name

rc.run('close all')

# Align all models and maps to the reference residue [and map]
for pair in class_dict['class_pairs'][:]:
    rc.run('open ' + reference_map_path)
    rc.run('open ' + reference_model_path)
    mrc_file = os.path.join(inp_dir, pair[0])
    pdb_file = os.path.join(inp_dir, pair[1])
    rc.run('open ' + mrc_file)
    rc.run('open ' + pdb_file)
    if res_type.upper() in align_atoms_dict['asymmetric'].keys():
        atoms_list = align_atoms_dict['asymmetric'][res_type.upper()]
        align_atoms = ','.join(atoms_list)
        rc.run(f'align #4@{align_atoms} to #2@{align_atoms}')
    else:
        atoms_list1 = align_atoms_dict['symmetric'][res_type.upper()][0]
        align_atoms1 = ','.join(atoms_list1)
        rms1 = rc.run(f'align #4@{align_atoms1} to #2@{align_atoms1}')[0][2]
        atoms_list2 = align_atoms_dict['symmetric'][res_type.upper()][1]
        align_atoms2 = ','.join(atoms_list2)
        rms2 = rc.run(f'align #4@{align_atoms2} to #2@{align_atoms2}')[0][2]
        if rms1 < rms2:
            rc.run(f'align #4@{align_atoms1} to #2@{align_atoms1}')

    rc.run('view position #3 sameAsModels #4')
    if fit_map:
        rc.run('volume #3 sdLevel 4')
        rc.run('fitmap #3 inMap #1 metric correlation')
    else:
        out_model = os.path.join(out_dir, pair[1])
        rc.run('save {} format mmcif models #4'.format(out_model))
    rc.run('vop resample  #3 onGrid #1')
    out_map = os.path.join(out_dir, pair[0])
    rc.run('save {} models #5'.format(out_map))
    # rc.run('close #3 #4 #5')
    # If just close the maps Chimera keeps the fitmap results in the memory (~1.GB per 10000 fits)
    # Close session deletes the results
    rc.run('close session')

rc.run('close all')
# name = class_dict['reference_model'].split('_')[0] + '_' + class_dict['class'] + '_representative.cif'
# reference_out = os.path.join(inp_dir, class_dict['class'], name)
# copy2(reference_model_path, reference_out)
rc.run('exit')



"""
mapMotifValidation.py

This module runs map versus amino acid residues map motif library validation.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import os
import sys
import csv
import json
import subprocess
import numpy as np
import psutil
import operator
import shutil
import logging
from datetime import datetime
import time
import argparse
from multiprocessing import Process, Manager
from distutils.spawn import find_executable
from Bio.PDB.Polypeptide import PPBuilder
# from strudel.chop.chopMap import ChopMap
from threed_strudel.utils import functions as func
import threed_strudel.configure as config
from threed_strudel.utils import bioUtils
from threed_strudel import nomenclature
from threed_strudel.chop.segmentResidueDensity import ExtractDensity

log = logging.getLogger(__name__)


class DictKeys:
    """
    Class for storing the fields names in the output csv file
    """
    RES_TYPE = 'residue_type'
    RES_NR = 'residue_nr'
    CHAIN = 'chain'
    RES_MAP = 'residue_map'
    SCORES = 'scores'
    #scores dict
    TYPE = 'type'
    CORR = 'corr'
    MAP = 'map'
    MATRIX = 'matrix'


class ComputeScores:
    """
    Class for correlations calculation between each amino acid residue map and each map motif
    """
    def __init__(self):
        self.residues = nomenclature.AA_RESIDUES_LIST
        self.aa_3to1 = nomenclature.AA_3LET_TO_1LET
        self.aa_1to3 = {v: k for k, v in self.aa_3to1.items()}
        self.chimera_path = self.find_chimerax()
        basedir = os.path.dirname(os.path.abspath(__file__))
        self.score_chimera_script = os.path.join(basedir, 'score_chimerax.py')
        self.segmented_pairs = 'segments_list.json'
        self.scores_file = 'scores.json'
        self.work_grid_sampling = 0.25
        self.chopped_pairs = Manager().list()
        # self.chop = ChopMap()
        self.extract = None
        self.segments = None
        self.work_dir = None
        self.lib = None
        self.in_map = None
        self.in_model = None
        self.input = None
        self.out_dir = None
        self.all_libs = '/Volumes/data/libs/libs'
        self.scores_paths = {}
        self.sequence_dir = None
        self.sequence_dict = {}
        self.keys = DictKeys()

    def find_chimerax(self):
        path = config.CHIMERA_PATH
        out = find_executable(path)
        if out is None:
            raise Exception(f'Could not find ChimeraX\n please edit {os.path.abspath(config.__file__)}')
        else:
            return path

    def set_dir_tree(self, work_dir):
        self.work_dir = os.path.abspath(work_dir)
        self.segments = os.path.join(self.work_dir, 'segments')
        self.input = os.path.join(self.work_dir, 'input')
        self.sequence_dir = os.path.join(self.work_dir, 'sequence')

        for path in [self.segments, self.input, self.out_dir, self.sequence_dir]:
            try:
                os.makedirs(path)
            except FileExistsError:
                pass
            except TypeError:
                pass

    def set_paths(self, work_dir, in_map, in_model, lib=None):
        """
        Creates the directory tree for the output and sets the input files paths
        :param work_dir: output directory
        :param lib: map motif library path
        :param in_map: input map path
        :param in_model: input model path
        """

        # print("WDD",self.work_dir)

        if os.path.exists(lib):
            self.lib = os.path.abspath(lib).rstrip('/')
            self.out_dir = os.path.join(self.work_dir, 'vs_' + os.path.basename(self.lib))
            self.lib = os.path.join(self.lib, 'motifs')
            self.check_motif_lib(self.lib)
        else:
            log.info('Motif library %s not found', lib)
        self.in_map = os.path.abspath(in_map)
        self.in_model = os.path.abspath(in_model)
        try:
            os.makedirs(self.out_dir)
        except FileExistsError:
            pass

        shutil.copy(self.in_map, self.input)
        shutil.copy(self.in_model, self.input)

    def set_results_paths(self, project_dir):
        folders = os.listdir(project_dir)

        tmp = os.path.join(project_dir, 'segments')
        if os.path.exists(tmp):
            self.segments = tmp
        for folder in folders:
            if folder.startswith('vs_'):
                path = os.path.join(project_dir, folder, 'scores.json')
                if os.path.exists(path):
                    lib = folder.split('_')[-1]
                    self.scores_paths[lib] = path
        self.all_libs = list(sorted(self.scores_paths.keys()))
        inp = os.path.join(project_dir, 'input')
        if os.path.exists(inp):
            in_files = os.listdir(inp)
            for ext in ['.cif', '.mmcif', '.pdb', '.ent']:
                for f in in_files:
                    if f.endswith(ext):
                        self.in_model = os.path.join(inp, f)
                        break
            for ext in ['.map', '.mrc']:
                for f in in_files:
                    if f.endswith(ext):
                        self.in_map = os.path.join(inp, f)
                        break
        seq_dir = os.path.join(project_dir, 'sequence')
        if os.path.exists(seq_dir):
            in_files = os.listdir(seq_dir)
            for f in in_files:
                if not f.startswith('.'):
                    seq_tmp = self.load_sequences(os.path.join(seq_dir, f))
                    if len(seq_tmp.keys()) > 0:
                        for k, v in seq_tmp.items():
                            self.sequence_dict[k] = v
        else:
            os.makedirs(seq_dir)
            self.sequence_dir = seq_dir

    @staticmethod
    def check_motif_lib(motif_lib_dir):
        """
        Checks the integrity of the motif library
        :param motif_lib_dir: motif library directory
        """
        log.info(f'Checking @{os.path.basename(motif_lib_dir)} motif library')

        files = os.listdir(motif_lib_dir)
        if len(files) == 0:
            log.error('Empty motif library')
        pdb_names = [i.split('.')[0] for i in files if i.endswith('.cif') and not i.startswith('.')]
        map_names = [i.split('.')[0] for i in files if (i.endswith('.mrc') or
                                                        i.endswith('.map')) and not i.startswith('.')]
        pdb_names.sort()
        map_names.sort()
        if len(pdb_names) != len(map_names):
            log.warning('There are missing files in the map motif library: %s', motif_lib_dir)
        else:
            log.info(f'OK')
        for name in pdb_names:
            if name not in map_names:
                log.warning('Missing map file for %s motif', name)
        for name in map_names:
            if name not in pdb_names:
                log.warning('Missing model file for %s motif', name)

    def segment_structure(self, n_cores=4, voxel=0.2, replace=True):
        """
        Chops the input map and model into amino-acid fragments
        :param n_cores: number of computing cores
        :param voxel: library voxel size
        :param replace: replace or not fragments chopped in during previous runs
        :return: segments list file path (json format)
        """
        # try:
        #     in_model = bioUtils.load_structure(self.in_model)[0]
        #     log.info('Failed to load structure using "auth" cif records. Attempting to use "label" records')
        # except:
        #     in_model = bioUtils.load_structure_label_id(self.in_model)[0]

        # del_atom_list = []
        # for chain in in_model:
        #     for residue in chain:
        #         for atom in residue:
        #             if atom.get_name().upper().startswith('H'):
        #                 del_atom_list.append((chain.get_id(), residue.get_id(), atom.get_id()))
        #
        # for chain, res, atom in del_atom_list:
        #     del in_model[chain][res][atom]

        segments_list_path = os.path.join(self.segments, self.segmented_pairs)
        if os.path.exists(segments_list_path):
            return segments_list_path

        self.extract = ExtractDensity(map_path=self.in_map, model_path=self.in_model, work_voxel_size=voxel)
        residues = [r for r in self.extract.in_model.get_residues() if r.get_resname().upper() in self.residues]
        # residues = residues[:1]
        log.info('Found %s residues in the input model', len(residues))
        log.info('Running map chopping using %s cores', n_cores)
        # map_obj_list, shifts_list = self.chop.chop_cube_list(residues, self.in_map, 4, zero_origin=False)

        split_residues = func.split_list(residues, n_cores)
        # split_map_obj_list = func.split_list(map_obj_list, n_cores)
        # split_shifts_list = func.split_list(shifts_list, n_cores)
        thread_list = []
        for i in range(len(split_residues)):
            t = Process(target=self._slave,
                        args=(split_residues[i], self.segments, replace))
            thread_list.append(t)
        for thread in thread_list:
            thread.start()
        for thread in thread_list:
            thread.join()

        out_pairs = []
        for i in self.chopped_pairs:
            out_pairs.append(i)

        segments_list_path = os.path.join(self.segments, self.segmented_pairs)
        if not os.path.exists(segments_list_path):
            func.save_json(segments_list_path, out_pairs)
        else:
            lst = func.read_json(segments_list_path)
            if len(out_pairs) > len(lst):
                p = psutil.Process()
                new_lst = os.path.join(self.segments, 'segments_' + str(p.pid) + '.json')
                func.save_json(new_lst, out_pairs)
                return new_lst

        return segments_list_path

    def _slave(self, residues, work_dir, replace):
        """
        Chopping worker
        :param residues: list of biopython residues objects
        :param map_obj_list: strudel map objects list
        :param shifts_list: list of translation matrices
        :param whole_model: input model as biopython object
        :param work_dir: work directory
        :param voxel: voxel size
        :param replace: replace or not fragments chopped in during previous runs
        """

        for residue in residues:
            res_type = residue.get_resname().lower()
            chain_id = residue.parent.id
            res_nr = residue.id[1]
            name_prefix = '{}-{}-{}'.format(res_type, res_nr, chain_id)

            res_cube_map_path = os.path.join(work_dir, name_prefix + '_cube_res.mrc')
            res_trace_map_path = os.path.join(work_dir, name_prefix + '_trace_res.mrc')
            res_walk_map_path = os.path.join(work_dir, name_prefix + '_walk_res.mrc')
            fin_map_path = os.path.join(work_dir, name_prefix + '_soft.mrc')

            if not os.path.exists(fin_map_path) or replace:
                ca_map, trace_map, walk_map, residue_map = self.extract.segment_residue_density(residue)
                if np.isnan(np.sum(residue_map.data)):
                    log.error("NaN values in {}".format(fin_map_path))
                residue_map.write_map(fin_map_path)
                ca_map.write_map(res_cube_map_path)
                trace_map.write_map(res_trace_map_path)
                walk_map.write_map(res_walk_map_path)
            self.chopped_pairs.append(fin_map_path)

    def compute_correlations(self, pairs_json, sd_level, n_cores=4, recompute=False, verbose=False):
        """
        Calculates input map fragments versus library map motifs correlations
        Runs Chimera as subprocess
        :param pairs_json: json file with chopped maps amd models paths
        :param n_cores: number of cores
        :param recompute: recompute or not known correlations from previous runs
        :param verbose: to run ChimeraX in verbose mode or not
        :return: output json file path
        """
        log.info('Running correlations calculations with ChimeraX')
        # Update keys in the chimeraX script
        updated_script = self.update_chimerax_script(self.score_chimera_script)
        updated_script = self.score_chimera_script
        if recompute:
            recompute = '-r'
        else:
            recompute = ''

        if verbose:
            v_flag = ''
        else:
            v_flag = '--silent'

        pairs_json = os.path.abspath(pairs_json)
        json_out = os.path.join(self.out_dir, self.scores_file)
        json_out = os.path.abspath(json_out)

        if sd_level:
            sd = f'-sd {sd_level}'
        else:
            sd = ''

        command = f'{self.chimera_path} --nogui {v_flag} {updated_script}' \
                  f' -s {pairs_json} -m {self.in_model} -l {self.lib} -np {n_cores} {sd} -o {json_out} {recompute} > chimera.log'

        subprocess.call(command, cwd=self.out_dir, shell=True)
        return json_out

    @staticmethod
    def update_chimerax_script(script_path):
        """
        Updates the dictionary keys in the Chimerax script
        :param script_path: Chimerax script path
        :return:
        """
        dd = DictKeys.__dict__
        with open(script_path, 'r') as f:
            lines = f.readlines()
        for key, value in dd.items():
            if not key.startswith('__'):
                for i, line in enumerate(lines):
                    if line.startswith('    ' + key):
                        lines[i] = f"    {key} = '{value}'\n"
        head, tail = os.path.split(script_path)
        updated_script_path = os.path.join(head, 'updated_' + tail)
        with open(updated_script_path, 'w') as of:
            for line in lines:
                of.write(line)
        return updated_script_path

    def score_sequence(self, seq, score_list, cut_off):
        if len(seq) != len(score_list):
            raise Exception("The length of the sequence_dir should be equal to the number of residues in the scores list")
        seq_score = 0
        for nr, char in enumerate(seq):
            try:
                code = self.aa_1to3[char.upper()].lower()
            except KeyError:
                continue
            try:
                corr = score_list[nr][self.keys.SCORES][code][1]
            except KeyError:
                print(f'Code {code} not valid')
                corr = 0
            # all_c = [i[1] for i in score_list[nr][self.keys.SCORES].values()]
            # quart = np.percentile(all_c, 80)
            # aver = sum(all_c)/len(all_c)
            if corr < cut_off:
            # if corr < quart:
                continue
            else:
                seq_score += corr
        # print(f'SEQ: {seq} SCORE: {seq_score}')
        return seq_score

    def score_sequence_(self, seq, score_list, cut_off=0.9):
        hgh = 'FTYWPHRL'
        if len(seq) != len(score_list):
            raise Exception("The length of the sequence_dir should be equal to the number of residues in the scores list")
        seq_score = 0
        for nr, char in enumerate(seq):
            try:
                code = self.aa_1to3[char.upper()].lower()
            except KeyError:
                continue
            try:
                corr = score_list[nr][self.keys.SCORES][code+'-class'][1]
            except KeyError:
                corr = 0
            all_c = [i[1] for i in score_list[nr][self.keys.SCORES].values()]
            aver = sum(all_c)/len(all_c)
            if not char.upper() in hgh:
                corr = corr * 0.9

            seq_score += corr

        return seq_score

    @staticmethod
    def load_sequences(file_path):
        with open(file_path, 'r') as fileobject:
            lines = fileobject.readlines()
        sequences = {}
        sequence = ''
        name = ''
        for line in lines:
            if not name and line[0] == '>':
                name = line[1:].rstrip()
            elif name and line[0] == '>':
                sequences[name] = sequence
                name = line[1:].rstrip()
                sequence = ''
            else:
                for char in line:
                    if char != '\n':
                        sequence += char
        sequences[name] = sequence
        return sequences

    @staticmethod
    def fragment_sequence(sequence, length):
        seq_lst = []
        for i in range(len(sequence) - length + 1):
            seq_lst.append(sequence[i:length + i])
        return seq_lst

    @staticmethod
    def get_sequence(model_path):
        model = bioUtils.load_structure(model_path)[0]
        ppb = PPBuilder()
        out_pp = []
        for pp in ppb.build_peptides(model):
            out_pp.append(pp.get_sequence())
        return out_pp

    def get_cutt_off(self, score_lst, percentile):
        all_c = []
        for nr in range(len(score_lst)):
            all_c += [i[1] for i in score_lst[nr][self.keys.SCORES].values()]
        print('ALL corr', len(all_c))
        return np.percentile(all_c, percentile)


    def test(self, score_lst, cs, cut_off):

        print("Sequence length: {}".format(len(score_lst)))
        rev_score_lst = score_lst[-1::-1]

        seq = self.sequence_dict
        # print(seq)
        keys = seq.keys()
        # print(keys)
        # print('Correct sequence: {}'.format(seq[keys[0]][start:end]))
        print('\nTop score sequences:')
        seq_scores = []
        length = len(score_lst)
        nr_s = 0
        tot_res = 0
        for full_seq in seq.values():
            # print(full_seq)
            nr_s += 1
            tot_res += len(full_seq)
            fragments = self.fragment_sequence(full_seq, length)
            # fragments.append('YAGDNFVRYTGDTISMAQTQLFAWEA')
            for fragment in fragments:
                f_score = self.score_sequence(fragment, score_lst, cut_off=cut_off)
                seq_scores.append((f_score, fragment, 'Forward'))
                f_score = self.score_sequence(fragment, rev_score_lst, cut_off=cut_off)
                seq_scores.append((f_score, fragment, 'Reversed'))

        seq_scores.sort(key=lambda z: z[0], reverse=True)
        # print("Total Sequences: {}".format(len(seq_scores)))
        print(f'Total sequences: {len(seq.values())}. Total fragments: {len(seq_scores)}')
        z = []
        for i, pair in enumerate(seq_scores[:]):

            scores = np.array([x[0] for x in seq_scores[:i] + seq_scores[i + 1:]])
            aver = scores.sum() / len(scores)

            Z = (pair[0] - aver) / scores.std()
            z.append(Z)
            # cs = 'YAGDNFVRYTGDTISMAQTQLFAWEA'
            # cs = 'LSRPERPDLVFEEEDLPYEEEIMRNQ'
            if pair[1] == cs:
                print('Correct sequencs:', cs)
                print('CORRECT Z= {}'.format(Z))

            if i < 10:
                seq_code = ''
                for key, value in seq.items():
                    if pair[1] in value:
                        seq_code = key
                print('Score= {:.4f}, Z= {:.4f}, Fragment: {} Parent seq name: {}, Direction: {}'.format(pair[0], Z,
                                                                                                         pair[1],
                                                                                                         seq_code,
                                                                                                         pair[2]))
            elif i > 1000:
                break
        all_scores = np.array([x[0] for x in seq_scores])
        aver = all_scores.sum() / len(all_scores)

        z = np.array(z)
        aver_z = z.sum() / len(z)

        print('Average score: {}'.format(aver))
        print('Average Z score: {}'.format(aver_z))
        print(nr_s)
        print(tot_res)

    def find_fragment(self, fragment_scores_lst, cut_off, correct_seq=None, verbose=True):
        # cut_off = self.get_cutt_off(fragment_scores_lst, 75)
        length = len(fragment_scores_lst)
        if verbose:
            print(f'Correct sequence: {correct_seq}')
            print("Sequence length: {}".format(len(fragment_scores_lst)))

        # rev_score_lst = fragment_scores_lst[-1::-1]
        seq = self.sequence_dict

        seq_scores = []
        nr_s = 0
        tot_res = 0
        for full_seq in seq.values():
            nr_s += 1
            tot_res += len(full_seq)
            fragments = self.fragment_sequence(full_seq, length)
            for fragment in fragments:
                f_score = self.score_sequence(fragment, fragment_scores_lst, cut_off=cut_off)
                seq_scores.append((f_score, fragment, 'Forward'))
                # f_score = self.score_sequence(fragment, rev_score_lst, cut_off=cut_off)
                # seq_scores.append((f_score, fragment, 'Reversed'))

        seq_scores.sort(key=lambda z: z[0], reverse=True)
        if verbose:
            print(f'Total sequences: {len(seq.values())}. Total fragments: {len(seq_scores)}')

        place = -1
        cor_Z = 0
        sc = 0
        for i, pair in enumerate(seq_scores[:]):

            scores = np.array([x[0] for x in seq_scores[:i] + seq_scores[i + 1:]])
            aver = scores.sum() / len(scores)

            Z = (pair[0] - aver) / scores.std()

            if i < 1000:
                seq_code = ''
                for key, value in seq.items():
                    if pair[1] in value:
                        seq_code = key
                if verbose:
                    print('Score= {:.4f}, Z= {:.4f}, Fragment: {} Parent seq name: {}, Direction: {}'.format(pair[0], Z,
                                                                                                         pair[1],
                                                                                                         seq_code,
                                                                                                         pair[2]))
                if pair[1] == correct_seq and place == -1:
                    place = i
                    cor_Z = Z
                    sc = pair[0]
            else:
                break

        return place, cor_Z, sc

    def find_fragment_per_chain_not_finished(self, fragment_scores_lst, cut_off, correct_seq=None, verbose=True):
        # cut_off = self.get_cutt_off(fragment_scores_lst, 75)
        length = len(fragment_scores_lst)
        if verbose:
            print(f'Correct sequence: {correct_seq}')
            print("Sequence length: {}".format(len(fragment_scores_lst)))

        rev_score_lst = fragment_scores_lst[-1::-1]
        seq = self.sequence_dict

        place = -1
        cor_Z = 0
        sc = 0
        for full_seq in seq.values():
            seq_scores = []

            fragments = self.fragment_sequence(full_seq, length)
            for fragment in fragments:
                f_score = self.score_sequence(fragment, fragment_scores_lst, cut_off=cut_off)
                seq_scores.append((f_score, fragment, 'Forward'))
                f_score = self.score_sequence(fragment, rev_score_lst, cut_off=cut_off)
                seq_scores.append((f_score, fragment, 'Reversed'))

            seq_scores.sort(key=lambda z: z[0], reverse=True)
            if verbose:
                print(f'Total sequences: {len(seq.values())}. Total fragments: {len(seq_scores)}')


            for i, pair in enumerate(seq_scores[:]):

                scores = np.array([x[0] for x in seq_scores[:i] + seq_scores[i + 1:]])
                aver = scores.sum() / len(scores)

                Z = (pair[0] - aver) / scores.std()

                if i < 1000:
                    seq_code = ''
                    for key, value in seq.items():
                        if pair[1] in value:
                            seq_code = key
                    if verbose:
                        print('Score= {:.4f}, Z= {:.4f}, Fragment: {} Parent seq name: {}, Direction: {}'.format(pair[0], Z,
                                                                                                             pair[1],
                                                                                                            seq_code,
                                                                                                           pair[2]))
                    if pair[1] == correct_seq and place == -1:
                        place = i
                        cor_Z = Z
                        sc = pair[0]
                else:
                    break

        return place, cor_Z, sc

    def score_fragments_old(self, scores_lst, length, cut_off, verbose=True):
        all_frag = []
        # c_off = self.get_cutt_off(scores_lst, 75)
        transl = nomenclature.AA_3LET_TO_1LET
        l = len(scores_lst)

        for i in range(l - length):
            seq = ''
            tmp = scores_lst[i:i+length]
            for e in tmp:
                code = transl[e["residue_type"].upper()]
                seq += code
            place, Z, sc = self.find_fragment(tmp, cut_off, seq, verbose)
            all_frag.append({'place': place, 'z': Z, 'seq': seq})
            # print(place, Z, seq, sc)
        return all_frag

    def find_record(self, score_list, chain, r_nr):
        for record in score_list:
            if record['chain'] == chain and record['residue_nr'] == r_nr:
                return record


    def score_fragments(self, scores_lst, length, cut_off, model_path, verbose=True):
        model = bioUtils.load_structure(model_path)[0]
        transl = nomenclature.AA_3LET_TO_1LET
        all_frag = []
        not_found = ''
        for ch in model.get_chains():
            if len(ch) < length:
                continue
            else:
                l = len(ch)
                all_res = [r for r in ch.get_residues()]
                for i in range(l - length):
                    fragm_scores = []
                    for res in all_res[i:i + length]:
                        c = res.parent.id
                        r_nr = res.id[1]
                        rec = self.find_record(scores_lst, c, r_nr)
                        if rec is not None:
                            fragm_scores.append(rec)
                        else:
                            not_found = f'{c}-{r_nr}'
                            break
                    seq = ''
                    for e in fragm_scores:
                        code = transl[e["residue_type"].upper()]
                        seq += code
                    if len(fragm_scores) == length:
                        place, Z, sc = self.find_fragment(fragm_scores, cut_off, seq, verbose)
                        all_frag.append({'place': place, 'z': Z, 'seq': seq, 'not_found': None})
                    else:
                        all_frag.append({'place': -1, 'z': 0, 'seq': '', 'not_found': not_found})

        return all_frag


def main():
    parser = argparse.ArgumentParser(description='Map Validation')
    parser.add_argument("-M", "--mode", dest="mode", required=True, help="Running mode")
    parser.add_argument("-s", "--sequence_dir", dest="seq", required=False, help="Atomic model path")
    parser.add_argument("-p", "--model", dest="in_model", required=True, help="Atomic model path")
    parser.add_argument("-m", "--map", dest="in_map", required=False, help="Map file path")
    parser.add_argument("-l", "--lib", dest="lib", required=False, help="Strudel motif library path")
    parser.add_argument("-np", "--n_processors", dest="np", required=False, default=2, help="Number of processors")
    parser.add_argument("-o", "--out", dest="out", required=True, help="Output directory")
    parser.add_argument("-log", "--log", dest="log", default=None, required=False, help="Log file")
    parser.add_argument("-v", "--voxel", dest="voxel", required=False, default=0.25, type=float, help="Segments voxel size")
    parser.add_argument("-r", "--recompute", dest="recompute_scores", action='store_true',
                        help="Recalculate correlations")
    parser.add_argument("-rs", "--recompute_segments", dest="recompute_segments", action='store_true',
                        help="Repeat segmenting of the input model and map if found in the output")
    parser.add_argument("-vc", "--verbose_chimerax", dest="v_c", action='store_true',
                        help="Run ChimeraX in verbose mode")
    parser.add_argument("-wl", "--warning_level", dest="warning_level", default='info',
                        help="Log file warning level [info[debug]")
    parser.add_argument("-sd", "--sd_level", dest="sd_level", required=False,
                        help="SD level for correlations calculations")
    parser.add_argument("-co", "--cut_off", dest="cut_off", type=int, required=False,
                        help="Cut-off quantile")


    args = parser.parse_args()
    if args.log:
        log_file = args.log
    else:
        log_file = os.path.join(args.out, 'sequence_finder.log')

    if not os.path.exists(args.out):
        os.makedirs(args.out)
    # print(args.out)

    logging.basicConfig(filename=log_file, level=logging.INFO, format='%(levelname)s:  %(message)s')
    date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
    logging.info('\n{:_^100}'.format('SequenceFinder') + '\n\nStarted: {}\n'.format(date_time))

    args.np = int(args.np)
    start = time.time()

    compute = ComputeScores()
    compute.set_dir_tree(work_dir=args.out)
    if args.lib:
        lib = args.lib
    else:
        lib = None

    if args.mode == 'C':
        compute.set_paths(work_dir=args.out, in_map=args.in_map, in_model=args.in_model, lib=lib)
        chopped_pairs = compute.segment_structure(n_cores=args.np, voxel=args.voxel, replace=args.recompute_segments)
        # chopped_pairs = os.path.join(args.out, 'segments', 'segments_list.json')
        json_file_path = compute.compute_correlations(chopped_pairs, args.sd_level, args.np, args.recompute_scores, verbose=args.v_c)
        #
        # prefix = os.path.splitext(json_file_path)[0]
    elif args.mode == 'F':
        # print(args.seq)
        if args.seq and os.path.exists(args.seq):
            shutil.copy(args.seq, compute.sequence_dir)
        compute.set_results_paths(args.out)
        # print("CORRECT SEQUENCE\n")
        # seq = compute.get_sequence(compute.in_model)
        # print(seq)
        for l, p in compute.scores_paths.items():
            scores_lst = func.read_json(p)
            tmp = []
            count = []
            for entry in scores_lst:
                if entry['residue_nr'] not in count:
                    tmp .append(entry)
                    count.append(entry['residue_nr'])
            # print("COUNTER", count)
            tmp = tmp[:]

            seq = ''
            transl = nomenclature.AA_3LET_TO_1LET
            for e in tmp:
                code = transl[e["residue_type"].upper()]
                seq += code
            print('Correct sequencs:', seq)
            c_off = compute.get_cutt_off(tmp, 75)
            print(c_off)
            # compute.test(tmp1, seq, cut_off=c_off)

            # result = compute.find_fragment(tmp[10:25], c_off, seq, True)
            # print(result)
            # o_j_p = '/Volumes/data/Work/test_segm/out_6/vs_2.5-3_formated/res_jsons_1000_60'
            # o_j_p = '/Volumes/data/find_seq_20584/out/vs_motifs_3.2-3.5/jsons_75'
            # o_j_p = '/Volumes/data/find_seq/untitled_folder/0711_out_sd1/jsons_75'
            compute.sequence_dict['scored'] = seq
            o_j_p = os.path.join(args.out, f'jsons_{args.cut_off}')
            if not os.path.exists(o_j_p):
                os.makedirs(o_j_p)
            for i in range(5, 40):
                o_j = os.path.join(o_j_p, f'{i}.json')
                res = compute.score_fragments(tmp, i, c_off, args.in_model, False)
                func.save_json(o_j, res)

                r_sc = [1 for i in res if i['place'] == 0]
                print(len(r_sc)/len(res) * 100)




    logging.info('\nElapsed: {}\n{:_^100}'.format(func.report_elapsed(start), ''))


if __name__ == '__main__':
    main()


from distutils.spawn import find_executable
import os

# ChimeraX executable path
# CHIMERA_PATH = '/nfs/msd/work2/andrei/soft/chimerax/18-12/redhat/opt/UCSF/ChimeraX-daily/bin/ChimeraX'
CHIMERA_PATH = '/Applications/ChimeraX-1.0.app/Contents/MacOS/ChimeraX'

out = find_executable(CHIMERA_PATH)
if out is None:
    print('WARNING: could not find ChimeraX\n please run:\n'
          'strudel_setChimeraX.py path_to_ChimeraX_executable')


"""
mapMotifValidation.py

This module runs map versus amino acid residues map motif library validation.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import os
import sys
import csv
import json
import subprocess
import numpy as np
import psutil
import operator
import shutil
import logging
from datetime import datetime
import time
import argparse
from multiprocessing import Process, Manager
from distutils.spawn import find_executable

from threed_strudel.chop.chopMap import ChopMap
from threed_strudel.utils import functions as func
import threed_strudel.configure as config
from threed_strudel.utils import bioUtils
from threed_strudel import nomenclature

log = logging.getLogger(__name__)


class DictKeys:
    """
    Class for storing the fields names in the output csv file
    """
    RES_TYPE = 'residue_type'
    RES_NR = 'residue_nr'
    CHAIN = 'chain'
    RES_MAP = 'residue_map'
    RES_MODEL = 'residue_model'
    M_TOP_TYPE = 'top_motif_type'
    M_TOP_NAME = 'top_motif_name'
    M_TOP_MATRIX = 'top_motif_matrix'
    TOP_CC = 'top_rscc'
    SAME_TYPE_NAME = 'same_type_motif_name'
    SAME_TYPE_CC = 'same_type_motif_rscc'
    SAME_TYPE_MATRIX = 'same_type_motif_matrix'
    # FULL NAME EXAMPLE "ala_cc" suffix for top correlation of specific residue type library motifs
    TYPE_TOP_CC = '_top_motif_rscc'
    # FULL NAME EXAMPLE "ala_motif_name"
    TYPE_TOP_NAME = '_top_motif_name'
    TYPE_TOP_MATRIX = '_top_motif_matrix'
    SCORES = 'scores'
    ALL_TOP_MOTIFS = 'all_top_motifs'
    TOP_SAME_TYPE_DIFF = 'top_same_dif'
    COMPLETE_DATA = 'complete_data'


def dict_list_to_csv(dict_list, csv_path):
    """
    Dumps a list of dictionaries with identical keys to a csv file
    :param dict_list: list of dictionaries
    :param csv_path: path
    """
    with open(csv_path, mode='w') as csv_file:
        fieldnames = [key for key in dict_list[0].keys()]
        writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
        writer.writeheader()
        for item in dict_list:
            writer.writerow(item)


def json_to_full_csv(json_path, csv_path):
    """
    Converts validation json output file to csv format
    :param json_path: input json file path
    :param csv_path: output csv path
    """
    k = DictKeys()
    with open(json_path) as j:
        data = json.load(j)

    out_data = []
    for item in data:
        tmp = {}
        for key, value in item.items():
            if key != k.SCORES:
                tmp[key] = value
            else:
                for el in value:
                    tmp[el[0]] = el[1]
        out_data.append(tmp)

    dict_list_to_csv(out_data, csv_path)


def csv_to_top_csv(csv_path, out_csv_path):
    """
    Prepares the input file for Strudel Score tool.
    Keeps only the top scoring motif data for each residue type
    :param csv_path: input scv file path
    :param out_csv_path: output scv file path
    """
    data_frame = []
    k = DictKeys()
    dd = DictKeys.__dict__
    k_values = [dd[k] for k in dd.keys() if not k.startswith('__')]

    with open(csv_path, mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file)

        for row in csv_reader:
            codes = nomenclature.AA_RESIDUES_LIST
            row_d = {}
            tmp = {}
            if any(['None' in v for v in row.values()]):
                row_d[k.COMPLETE_DATA] = False
            else:
                row_d[k.COMPLETE_DATA] = True
            for key, value in row.items():
                code = key[:3]
                if code.upper() in codes:
                    corr, _, matrix = value.partition('_m_')
                    if corr == 'None':
                        corr = None
                    if matrix == 'None':
                        matrix = None
                    if code not in tmp.keys():
                        if 'None' not in value:
                            tmp[code] = [float(corr), key, matrix]
                        else:
                            tmp[code] = [None, None, None]
                    elif tmp[code][0] is not None and 'None' not in value:
                        if tmp[code][0] < float(corr):
                            tmp[code][0] = float(corr)
                            tmp[code][1] = key
                            tmp[code][2] = matrix

                elif key in k_values:
                    try:
                        row_d[key] = int(value)
                    except ValueError:
                        row_d[key] = value
                else:
                    raise Exception(f'Key {key} is unknown to the data structure cannot continue')

            max_correlation = -1
            row_d[k.M_TOP_TYPE] = None
            row_d[k.M_TOP_NAME] = None
            row_d[k.TOP_CC] = None
            row_d[k.SAME_TYPE_NAME] = None
            row_d[k.SAME_TYPE_CC] = None
            row_d[k.SAME_TYPE_MATRIX] = None
            for key, value in tmp.items():
                if key == row_d[k.RES_TYPE]:
                    row_d[k.SAME_TYPE_NAME] = value[1]
                    row_d[k.SAME_TYPE_CC] = value[0]
                    row_d[k.SAME_TYPE_MATRIX] = value[2]
                if value[0] is not None:
                    if value[0] > max_correlation:
                        max_correlation = value[0]
                        row_d[k.M_TOP_TYPE] = key
                        row_d[k.M_TOP_NAME] = value[1]
                        row_d[k.TOP_CC] = value[0]
                        row_d[k.M_TOP_MATRIX] = value[2]

            for key, value in tmp.items():
                row_d[key + k.TYPE_TOP_CC] = value[0]
                row_d[key + k.TYPE_TOP_NAME] = value[1]
                row_d[key + k.TYPE_TOP_MATRIX] = value[2]
            data_frame.append(row_d)

    data_frame.sort(key=operator.itemgetter(k.CHAIN, k.RES_NR))
    dict_list_to_csv(data_frame, out_csv_path)


def csv_to_top_scores_only_csv(csv_path, out_csv_path):
    """
    Keeps only minimum data for the web version of the Strudel Score.
    :param csv_path: input scv file path
    :param out_csv_path: output scv file path
    """
    data_frame = []
    k = DictKeys()
    dd = DictKeys.__dict__
    k_values = [dd[k] for k in dd.keys() if not k.startswith('__')]

    with open(csv_path, mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file)

        for row in csv_reader:
            codes = nomenclature.AA_RESIDUES_LIST
            row_d = {}
            tmp = {}
            if any(['None' in v for v in row.values()]):
                row_d[k.COMPLETE_DATA] = False
            else:
                row_d[k.COMPLETE_DATA] = True
            for key, value in row.items():
                code = key[:3]
                if code.upper() in codes:
                    corr, _, matrix = value.partition('_m_')
                    if code not in tmp.keys():
                        if 'None' not in value:
                            tmp[code] = [float(corr), key]
                        else:
                            tmp[code] = [None, None]

                    elif tmp[code][0] is not None and 'None' not in value:
                        if tmp[code][0] < float(corr):
                            tmp[code][0] = float(corr)
                            tmp[code][1] = key

                elif key in k_values:
                    try:
                        row_d[key] = int(value)
                    except ValueError:
                        row_d[key] = value
                else:
                    raise Exception(f'Key {key} is unknown to the data structure cannot continue')

            max_correlation = -1
            row_d[k.M_TOP_TYPE] = None
            row_d[k.TOP_CC] = None
            for key, value in tmp.items():
                if key == row_d[k.RES_TYPE]:
                    row_d[k.SAME_TYPE_CC] = value[0]
                if value[0] is not None:
                    if value[0] > max_correlation:
                        max_correlation = round(value[0], 5)
                        row_d[k.M_TOP_TYPE] = key
                        row_d[k.TOP_CC] = round(value[0], 5)

            for key, value in tmp.items():
                row_d[key] = value[0]
            data_frame.append(row_d)
    data_frame.sort(key=operator.itemgetter(k.CHAIN, k.RES_NR))
    dict_list_to_csv(data_frame, out_csv_path)


class ComputeScores:
    """
    Class for correlations calculation between each amino acid residue map and each map motif
    """
    def __init__(self):
        self.residues = nomenclature.AA_RESIDUES_LIST
        self.chimera_path = self.find_chimerax()
        basedir = os.path.dirname(os.path.abspath(__file__))
        self.score_chimera_script = os.path.join(basedir, 'score_chimerax.py')
        self.segmented_pairs = 'segments_list.json'
        self.scores_file = 'scores.json'
        self.work_grid_sampling = 0.25
        self.chopped_pairs = Manager().list()
        self.chop = ChopMap()
        self.segments = None
        self.work_dir = None
        self.lib = None
        self.in_map = None
        self.in_model = None
        self.input = None
        self.out_dir = None

    def find_chimerax(self):
        path = config.CHIMERA_PATH
        out = find_executable(path)
        if out is None:
            raise Exception(f'Could not find ChimeraX\n please edit {os.path.abspath(config.__file__)}')
        else:
            return path

    def set_paths(self, work_dir, lib, in_map, in_model):
        """
        Creates the directory tree for the output and sets the input files paths
        :param work_dir: output directory
        :param lib: map motif library path
        :param in_map: input map path
        :param in_model: input model path
        """
        self.work_dir = os.path.abspath(work_dir)
        if os.path.exists(lib):
            self.lib = os.path.abspath(lib).rstrip('/')
            self.out_dir = os.path.join(self.work_dir, 'vs_' + os.path.basename(self.lib))
            self.lib = os.path.join(self.lib, 'motifs')
            self.check_motif_lib(self.lib)
        else:
            log.info('Motif library %s not found', lib)
        self.in_map = os.path.abspath(in_map)
        self.in_model = os.path.abspath(in_model)

        self.segments = os.path.join(self.work_dir, 'segments')
        self.input = os.path.join(self.work_dir, 'input')

        for path in [self.segments, self.input, self.out_dir]:
            try:
                os.makedirs(path)
            except FileExistsError:
                pass

        shutil.copy(self.in_map, self.input)
        shutil.copy(self.in_model, self.input)

    @staticmethod
    def check_motif_lib(motif_lib_dir):
        """
        Checks the integrity of the motif library
        :param motif_lib_dir: motif library directory
        """
        log.info(f'Checking @{os.path.basename(motif_lib_dir)} motif library')

        files = os.listdir(motif_lib_dir)
        if len(files) == 0:
            log.error('Empty motif library')
        pdb_names = [i.split('.')[0] for i in files if i.endswith('.cif') and not i.startswith('.')]
        map_names = [i.split('.')[0] for i in files if (i.endswith('.mrc') or
                                                        i.endswith('.map')) and not i.startswith('.')]
        pdb_names.sort()
        map_names.sort()
        if len(pdb_names) != len(map_names):
            log.warning('There are missing files in the map motif library: %s', motif_lib_dir)
        else:
            log.info(f'OK')
        for name in pdb_names:
            if name not in map_names:
                log.warning('Missing map file for %s motif', name)
        for name in map_names:
            if name not in pdb_names:
                log.warning('Missing model file for %s motif', name)

    def chop_structure(self, n_cores=4, voxel=0.25, replace=True):
        """
        Chops the input map and model into amino-acid fragments
        :param n_cores: number of computing cores
        :param voxel: library voxel size
        :param replace: replace or not fragments chopped in during previous runs
        :return: segments list file path (json format)
        """
        try:
            in_model = bioUtils.load_structure(self.in_model)[0]
            log.info('Failed to load structure using "auth" cif records. Attempting to use "label" records')
        except:
            in_model = bioUtils.load_structure_label_id(self.in_model)[0]

        del_atom_list = []
        for chain in in_model:
            for residue in chain:
                for atom in residue:
                    if atom.get_name().upper().startswith('H'):
                        del_atom_list.append((chain.get_id(), residue.get_id(), atom.get_id()))

        for chain, res, atom in del_atom_list:
            del in_model[chain][res][atom]

        residues = [r for r in in_model.get_residues() if r.get_resname().upper() in self.residues]
        # residues = residues[:]
        log.info('Found %s residues in the input model', len(residues))
        log.info('Running map chopping using %s cores', n_cores)
        map_obj_list, shifts_list = self.chop.chop_cube_list(residues, self.in_map, 4, zero_origin=False)
        split_residues = func.split_list(residues, n_cores)
        split_map_obj_list = func.split_list(map_obj_list, n_cores)
        split_shifts_list = func.split_list(shifts_list, n_cores)
        thread_list = []
        for i in range(len(split_residues)):
            t = Process(target=self._slave,
                        args=(split_residues[i], split_map_obj_list[i],
                              split_shifts_list[i], in_model, self.segments, voxel, replace))
            thread_list.append(t)
        for thread in thread_list:
            thread.start()
        for thread in thread_list:
            thread.join()

        out_pairs = []
        for i in self.chopped_pairs:
            out_pairs.append(i)

        segments_list_path = os.path.join(self.segments, self.segmented_pairs)
        if not os.path.exists(segments_list_path):
            func.save_json(segments_list_path, out_pairs)
        else:
            lst = func.read_json(segments_list_path)
            if len(out_pairs) > len(lst):
                p = psutil.Process()
                new_lst = os.path.join(self.segments, 'segments_' + str(p.pid) + '.json')
                func.save_json(new_lst, out_pairs)
                return new_lst

        return segments_list_path

    def _slave(self, residues, map_obj_list, shifts_list, whole_model, work_dir, voxel, replace):
        """
        Chopping worker
        :param residues: list of biopython residues objects
        :param map_obj_list: strudel map objects list
        :param shifts_list: list of translation matrices
        :param whole_model: input model as biopython object
        :param work_dir: work directory
        :param voxel: voxel size
        :param replace: replace or not fragments chopped in during previous runs
        """

        for i in range(len(residues)):
            residue = residues[i]
            matrix = shifts_list[i]

            res_type = residue.get_resname().lower()
            chain_id = residue.parent.id
            res_nr = residue.id[1]
            name_prefix = '{}-{}-{}'.format(res_type, res_nr, chain_id)

            # cube_map_path = os.path.join(work_dir, name_prefix + '_cube.mrc')
            res_cube_map_path = os.path.join(work_dir, name_prefix + '_cube_res.mrc')
            fin_map_path = os.path.join(work_dir, name_prefix + '_soft.mrc')
            residue_path = os.path.join(work_dir, name_prefix + '.cif')

            if not os.path.exists(fin_map_path) or not os.path.exists(residue_path) or replace:
                # map_obj_list[i].write_map(cube_map_path)
                cube_map_obj = map_obj_list[i]
                # cube_map_obj.write_map(cube_map_path)

                cube_map_obj.grid_resample_emda(voxel)

                # cube_map_obj.write_map(res_cube_map_path)

                bioUtils.shift_coord(matrix, residue)
                struct = bioUtils.residues2structure(residue)
                bioUtils.save_model(struct, residue_path)
                side_chain = bioUtils.del_main_chain(residue)
                # fin_map = self.chop.chop_soft_radius(side_chain, res_cube_map_path, hard_radius=2, soft_radius=1,)
                fin_map = self.chop.chop_soft_radius_watershed(side_chain, cube_map_obj, whole_model, radius=2, soft_radius=1, )
                if np.isnan(np.sum(fin_map.data)):
                    log.error("NaN values in {}".format(fin_map_path))
                fin_map.write_map(fin_map_path)

            self.chopped_pairs.append((fin_map_path, residue_path))

    def compute_correlations(self, pairs_json, n_cores=4, recompute=False, verbose=False):
        """
        Calculates input map fragments versus library map motifs correlations
        Runs Chimera as subprocess
        :param pairs_json: json file with chopped maps amd models paths
        :param n_cores: number of cores
        :param recompute: recompute or not known correlations from previous runs
        :param verbose: to run ChimeraX in verbose mode or not
        :return: output json file path
        """
        log.info('Running correlations calculations with ChimeraX')
        # Update keys in the chimeraX script
        updated_script = self.update_chimerax_script(self.score_chimera_script)
        if recompute:
            recompute = '-r'
        else:
            recompute = ''

        if verbose:
            v_flag = ''
        else:
            v_flag = '--silent'

        pairs_json = os.path.abspath(pairs_json)
        json_out = os.path.join(self.out_dir, self.scores_file)
        json_out = os.path.abspath(json_out)

        command = f'{self.chimera_path} --nogui {v_flag} {updated_script}' \
                  f' -p {pairs_json} -l {self.lib} -np {n_cores} -o {json_out} {recompute} > chimera.log'

        subprocess.call(command, cwd=self.out_dir, shell=True)
        return json_out

    @staticmethod
    def update_chimerax_script(script_path):
        """
        Updates the dictionary keys in the Chimerax script
        :param script_path: Chimerax script path
        :return:
        """
        dd = DictKeys.__dict__
        with open(script_path, 'r') as f:
            lines = f.readlines()
        for key, value in dd.items():
            if not key.startswith('__'):
                for i, line in enumerate(lines):
                    if line.startswith('    ' + key):
                        lines[i] = f"    {key} = '{value}'\n"
        head, tail = os.path.split(script_path)
        updated_script_path = os.path.join(head, 'updated_' + tail)
        with open(updated_script_path, 'w') as of:
            for line in lines:
                of.write(line)
        return updated_script_path


def main():
    parser = argparse.ArgumentParser(description='Map Validation')
    parser.add_argument("-p", "--model", dest="in_model", required=True, help="Atomic model path")
    parser.add_argument("-m", "--map", dest="in_map", required=True, help="Map file path")
    parser.add_argument("-l", "--lib", dest="lib", required=True, help="Strudel motif library path")
    parser.add_argument("-np", "--n_processors", dest="np", required=False, default=2, help="Number of processors")
    parser.add_argument("-o", "--out", dest="out", required=True, help="Output directory")
    parser.add_argument("-log", "--log", dest="log", default=None, required=False, help="Log file")
    parser.add_argument("-v", "--voxel", dest="voxel", required=False, default=0.25, type=float, help="Segments voxel size")
    parser.add_argument("-r", "--recompute", dest="recompute_scores", action='store_true',
                        help="Recalculate correlations")
    parser.add_argument("-rs", "--recompute_segments", dest="recompute_segments", action='store_true',
                        help="Repeat segmenting of the input model and map if found in the output")
    parser.add_argument("-vc", "--verbose_chimerax", dest="v_c", action='store_true',
                        help="Run ChimeraX in verbose mode")
    parser.add_argument("-wl", "--warning_level", dest="warning_level", default='info',
                        help="Log file warning level [info[debug]")

    args = parser.parse_args()
    if args.log:
        log_file = args.log
    else:
        log_file = os.path.join(args.out, 'map_motif_validation.log')

    if not os.path.exists(args.out):
        os.makedirs(args.out)

    logging.basicConfig(filename=log_file, level=logging.INFO, format='%(levelname)s:  %(message)s')
    date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
    logging.info('\n{:_^100}'.format('MapMotifValidation') + '\n\nStarted: {}\n'.format(date_time))

    args.np = int(args.np)
    start = time.time()

    compute = ComputeScores()
    compute.set_paths(work_dir=args.out, lib=args.lib, in_map=args.in_map, in_model=args.in_model)
    chopped_pairs = compute.chop_structure(n_cores=args.np, voxel=args.voxel, replace=args.recompute_segments)
    json_file_path = compute.compute_correlations(chopped_pairs, args.np, args.recompute_scores, verbose=args.v_c)

    prefix = os.path.splitext(json_file_path)[0]
    if os.path.exists(prefix + '.csv'):
        csv_to_top_csv(prefix+'.csv', prefix+'_top.csv')
        csv_to_top_scores_only_csv(prefix + '.csv', prefix + '_top_for_web.csv')

    logging.info('\nElapsed: {}\n{:_^100}'.format(func.report_elapsed(start), ''))


if __name__ == '__main__':
    main()


from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.MMCIFParser import MMCIFParser, FastMMCIFParser
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.mmcifio import MMCIFIO
from Bio.PDB.Polypeptide import PPBuilder
import sys
import json
import copy
import os
import argparse
import subprocess

AA_3LET_TO_1LET = {'ALA': 'A',
                   'ARG': 'R',
                   'ASN': 'N',
                   'ASP': 'D',
                   'CYS': 'C',
                   'GLU': 'E',
                   'GLN': 'Q',
                   'GLY': 'G',
                   'HIS': 'H',
                   'ILE': 'I',
                   'LEU': 'L',
                   'LYS': 'K',
                   'MET': 'M',
                   'PHE': 'F',
                   'PRO': 'P',
                   'SER': 'S',
                   'THR': 'T',
                   'TRP': 'W',
                   'TYR': 'Y',
                   'VAL': 'V'}


def load_structure(model_path):
    """
    Creates Biopython structure object from file
    :param model_path: pdb, cif file in_dir
    :return: Biopython structure object
    """
    def copy_label_seq_id(in_path):
        parser = MMCIFParser(QUIET=True)
        parser._mmcif_dict = MMCIF2Dict(in_path)
        if "_atom_site.auth_seq_id" in parser._mmcif_dict:
            for i in parser._mmcif_dict["_atom_site.auth_seq_id"]:
                try:
                    int(i)
                except ValueError:
                    parser._mmcif_dict["_atom_site.auth_seq_id"] = parser._mmcif_dict["_atom_site.label_seq_id"]
        parser._build_structure(in_path.split('/')[-1].split('.')[0])
        return parser._structure_builder.get_structure()

    if model_path.split('.')[-1] == 'pdb' or model_path.split('.')[-1] == 'ent':
        parser = PDBParser(PERMISSIVE=1, QUIET=True)
        structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
    elif model_path.split('.')[-1] == 'cif':
        try:
            parser = MMCIFParser(QUIET=True)
            structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
        except ValueError:
            structure = copy_label_seq_id(model_path)
    else:
        raise Exception('Please provide the input residue in pdb or cif format')

    return structure


def save_model(model, out_path, preserve_atom_numbering=False):
    """
    Save residue to file
    :param model: structure object
    :param out_path: output file in_dir
    """
    ext = out_path.split('.')[-1]
    if ext == 'pdb' or ext == 'ent':
        io = PDBIO()
    else:
        io = MMCIFIO()
    io.set_structure(model)
    io.save(out_path, preserve_atom_numbering=preserve_atom_numbering)


def make_poly_ala(model):
    del_atoms = []
    bb = ['CA', 'C', 'N', 'O', 'CB']
    for chain in model:
        for residue in chain:
            for atom in residue:
                if atom.get_name().upper() not in bb:
                    del_atoms.append((chain.get_id(), residue.get_id(), atom.get_id()))
    for chain, resi, atom in del_atoms:
        del model[chain][resi][atom]

    for chain in model:
        for residue in chain:
            residue.resname = 'ALA'
    return model


def get_fragments(model, length):
    fragments = []
    all_res = [a for a in model.get_residues()]
    ch = all_res[0].get_parent().get_id()
    l = len(all_res)
    for i in range(l-length):
        tmp_model = copy.deepcopy(model)
        del_ids = []
        for res in all_res:
            if res not in all_res[i:i+length]:
                del_ids.append(res.get_id())
        for _id in del_ids:
            del tmp_model[ch][_id]
        
        fragments.append(tmp_model)
    return fragments


def get_sequence(model):
    code = ''
    for residue in model.get_residues():
        r_c = residue.resname
        code += AA_3LET_TO_1LET[r_c.upper()]
    return code


def save_fragments(fragments_list, frag_path):
    for m in fragments_list:
        seq = get_sequence(m)
        rs = [r for r in m.get_residues()]
        st = rs[0].get_id()[1]
        end = rs[-1].get_id()[1]
        o_p = os.path.join(frag_path, f'{seq}_{st}-{end}.pdb')
        m = make_poly_ala(m)
        save_model(m, o_p)


def run_buc(buc_path, pdb_path, mtzfile, out_path, seq_path):

    command = f'''
{buc_path} -stdin << eof
seqin {seq_path}
mtzin {mtzfile}
colin-fo Fout0,SIGFem
colin-free FreeR_flag
colin-phifom Pout0,FOMem
pdbin {pdb_path}
cycles 1
sequence
eof
'''

    log = open(out_path, 'a')
    log.writelines(command)
    log.flush()
    proc = subprocess.Popen(command,
                            universal_newlines=True, stdout=log, shell=True)
    sys.stdout.flush()
    proc.wait()
    log.flush()
    log.close()


def read_buc_out(log_file, seq):
    with open(log_file, 'r') as f:
        lines = f.readlines()
        place, z = -1, 0
        for line in lines:
            words = line.split()
            if seq in words:
                place = int(words[0])
                z = float(words[1])
                break
    return place, z


def main():
    parser = argparse.ArgumentParser(description='Run Buccaneer')
    parser.add_argument("-buc", "--buccaneer", dest="buc_path", required=False, help="Buccaneer path")
    parser.add_argument("-f", "--fragments_dir", dest="f_dir", required=True, help="PDB fragments directory")
    parser.add_argument("-t", "--task", dest="task", required=True, help="Task to run. F - generate fragments"
                                                                          "R - running buccaneer")
    parser.add_argument("-j", "--json_out", dest="j_out", required=False, help="Json output file")
    parser.add_argument("-o", "--buc_out", dest="buc_out", required=False, help="Buccaneer output")
    parser.add_argument("-p", "--model", dest="m_path", required=True, help="Model path")
    parser.add_argument("-m", "--mtz", dest="mtz_path", required=False, help="Mtz path")
    parser.add_argument("-s", "--seq", dest="seq_path", required=False, help="Sequence path")
    parser.add_argument("-len", "--len", dest="len", required=False, type=int, help="Fragments length")

    args = parser.parse_args()

    for p in [args.f_dir, args.buc_out]:
        if p is not None:
            if not os.path.exists(p):
                os.makedirs(p)

    if args.task == 'F':
        model = load_structure(args.m_path)[0]
        frags = get_fragments(model, args.len)
        save_fragments(frags, args.f_dir)

    if args.task == 'R':
        frag_files = os.listdir(args.f_dir)
        pdb_files = [f for f in frag_files if f.endswith('.pdb') and not f.startswith('.')]
        if len(pdb_files) == 0:
            print('No PDB files found')
            sys.exit(0)
        else:
            results_list = []
            for pdb in pdb_files:
                seq = pdb.split('_')[0]
                lod_file = pdb.split('.')[0] + '.txt'
                log_path = os.path.join(args.buc_out, lod_file)
                pdb_path = os.path.join(args.f_dir, pdb)
                run_buc(args.buc_path, pdb_path, args.mtz_path, log_path, args.seq_path)
                place, z = read_buc_out(log_path, seq)
                results = {'place': place,
                           'z': z,
                           'seq': seq}
                results_list.append(results)
            
                if args.j_out:
                    with open(args.j_out, 'w') as j:
                        json.dump(results_list, j, indent=4)
                else:
                    print(f'{place} {z} {seq}')
                

if __name__ == '__main__':
    main()



"""
score_chimerax.py

This script calculates correlations

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""


__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2019-04-10'


import os
import csv
import sys
from multiprocessing import Process, Manager, Lock
from datetime import datetime
import multiprocessing
from shutil import copy, rmtree
import json
import argparse
import operator
import psutil
# from chimera import runCommand as rc
# from chimera import openModels as om, selection
# from FitMap.fitcmd import fitmap
from chimerax.core.commands import Command
from chimerax.map.volume import Volume
from chimerax.atomic.structure import AtomicStructure
import logging

rc = Command(session)
SD_LEVEL = 3


class DictKeys:
    """
    Class for storing the fields names in the output csv file
    """
    RES_TYPE = 'residue_type'
    RES_NR = 'residue_nr'
    CHAIN = 'chain'
    TOP_TYPE = 'top_residue_type'
    TOP_CC = 'top_cc'
    RES_MAP = 'residue_map'
    RES_MODEL = 'residue_model'
    SCORES = 'scores'


keys = DictKeys()

align_atoms_dict = {
    'symmetric': {'HIS': [['n', 'c', 'ca', 'cb', 'cg', 'nd1', 'ce1', 'cd2', 'ne2'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ne2', 'nd1', 'ce1']],
                  'PHE': [['n', 'c', 'ca', 'cb', 'cg', 'cd1', 'ce1', 'cd2', 'ce2', 'cz'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ce2', 'cd1', 'ce1', 'cz']],
                  'TYR': [['n', 'c', 'ca', 'cb', 'cg', 'cd1', 'ce1', 'cd2', 'ce2', 'cz'],
                          ['n', 'c', 'ca', 'cb', 'cg', 'cd2', 'ce2', 'cd1', 'ce1', 'cz']]
                  },
    'asymmetric': {'ASP': ['n', 'c', 'ca', 'cb', 'cg'],
                   'GLU': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'ASN': ['n', 'c', 'ca', 'cb', 'cg'],
                   'ARG': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'LYS': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'MET': ['n', 'c', 'ca', 'cb', 'cg', 'sd'],
                   'GLN': ['n', 'c', 'ca', 'cb', 'cg', 'cd'],
                   'ILE': ['n', 'c', 'ca', 'cb', 'cg1', 'cd1'],
                   'LEU': ['n', 'c', 'ca', 'cb', 'cg', 'cd1'],
                   'TRP': ['n', 'c', 'ca', 'cb', 'cg', 'cd1'],
                   'PRO': ['n', 'c', 'ca', 'cb', 'cg'],
                   'THR': ['n', 'c', 'ca', 'cb', 'cg2'],
                   'VAL': ['n', 'c', 'ca', 'cb', 'cg1', 'cg2'],
                   'SER': ['n', 'c', 'ca', 'cb', 'og'],
                   'CYS': ['n', 'c', 'ca', 'cb', 'sg'],
                   'GLY': ['n', 'c', 'ca'],
                   'ALA': ['n', 'c', 'ca', 'cb']
                   }
                    }


def run_x(command):
    try:
        return rc.run(command)[0][0]
    except TypeError:
        return None


def open_model(path, log):
    try:
        model_obj = session.open_command.open_data(path)[0][0]
        if type(model_obj) == AtomicStructure or type(model_obj) == Volume:
            session.models.add([model_obj])
            return model_obj
    except AttributeError:
        model_obj = run_x(f'open {path}')
        if type(model_obj) == AtomicStructure or type(model_obj) == Volume:
            return model_obj
        else:
            log.error(f'Could not get model object reference after opening:\n {path}')


def read_motif_lib(motif_lib_dir):
    """
    Reads motif library
    :param motif_lib_dir: library path
    :return: list of map model pair paths
    """
    lib_pairs = []
    files = os.listdir(motif_lib_dir)
    pdb_names = [i for i in files if i.endswith('.cif') and not i.startswith('.')]
    for name in pdb_names:
        prefix = name.split('.cif')[0]
        map_path = os.path.join(motif_lib_dir, prefix + '.map')
        mrc_path = os.path.join(motif_lib_dir, prefix + '.mrc')
        if os.path.exists(map_path):
            lib_pairs.append([prefix, os.path.join(motif_lib_dir, name), map_path])
        elif os.path.exists(mrc_path):
            lib_pairs.append([prefix, os.path.join(motif_lib_dir, name), mrc_path])
    return lib_pairs


def score_residue(res_map, res_model, loaded_lib, log):
    """
    Scores a residue map against the motif library
    :param res_map: residue map path
    :param res_model: residue model path
    :param loaded_lib: motif library as chimerax objects
    :param log: logger
    :return: residue scores
    """
    log.info(f'Scoring {os.path.basename(res_map)}, {os.path.basename(res_model)} ')
    res_score = []
    # mod = run_x('open ' + res_model)
    # vol = run_x('open ' + res_map)
    mod = open_model(res_model, log)
    vol = open_model(res_map, log)
    for motif, lib_mod, lib_vol in loaded_lib:
        mod_sel = '#{}@c,ca,n'.format(mod.id_string)
        lib_mod_sel = '#{}@c,ca,n'.format(lib_mod.id_string)
        try:
            run_x(f'align {lib_mod_sel} to {mod_sel}')
        except:
            log.error(f'Backbone atoms missing in {os.path.basename(res_model)}. Cannot align! ')
            res_score.append((motif, None))
            continue
        run_x(f'view position #{lib_vol.id_string} sameAsModels #{lib_mod.id_string}')
        run_x(f'volume #{lib_vol.id_string} sdLevel {SD_LEVEL}')
        run_x(f'volume #{vol.id_string} sdLevel {SD_LEVEL}')

        try:
            fit1 = run_x(f'fitmap #{lib_vol.id_string} inMap #{vol.id_string} metric correlation')
            correlation1 = fit1.correlation()
        except Exception as err:
            log.error('Chimerax user error!! %s in map %s at sdLevel =  %s', err, lib_vol.name, SD_LEVEL)
            log.info('Changing threshold level to %s', SD_LEVEL - 0.1)
            try:
                run_x(f'volume #{lib_vol.id_string} sdLevel {SD_LEVEL - 0.1}')
                fit1 = run_x(f'fitmap #{lib_vol.id_string} inMap #{vol.id_string} metric correlation')
                correlation1 = fit1.correlation()
            except Exception as err:
                log.error('Chimerax user error!! %s in map %s at sdLevel =  %s', err, lib_vol.name, SD_LEVEL-0.1)
                log.info('Could not calculate %s and %s correlation', vol.name, lib_vol.name)
                correlation1 = None
        matrix = ','.join(f"{x}" for x in tuple(lib_vol.position.matrix.flat))
        try:
            fit2 = run_x(f'fitmap #{vol.id_string} inMap #{lib_vol.id_string} metric correlation')
            correlation2 = fit2.correlation()
        except Exception as err:
            log.info('User error!! %s %s', err, lib_vol.name)
            try:
                run_x(f'volume #{vol.id_string} sdLevel {SD_LEVEL - 0.1}')
                fit2 = run_x(f'fitmap #{vol.id_string} inMap #{lib_vol.id_string} metric correlation')
                correlation2 = fit2.correlation()
            except Exception as err:
                log.info('User error!! %s %s', err, lib_vol.name)
                correlation2 = None
        if all([correlation1, correlation2]):
            min_corr = min([correlation1, correlation2])
            res_score.append((motif, min_corr, matrix))
        else:
            res_score.append((motif, None, None))
        # This resets the view after fitting
        run_x('view orient; view initial')

    # res_score.sort(reverse=True, key=lambda z: z[1])
    return res_score


def capture_memory_usage():
    """
    Finds the memory usage of the current process
    :return: memory usage in MB
    """
    p = psutil.Process()
    mem = psutil.Process(p.pid).memory_info()
    return p.pid, mem.rss / 1000000


def slave(pairs_list, lib, score_list, lock, json_out_path, csv_out_path=None, close_rate=50):
    p = psutil.Process()
    log = logging.getLogger(str(p.pid))

    logging.basicConfig(filename='process_logs/' + str(p.pid) + '.log', level=logging.INFO,
                        format='%(levelname)s:  %(message)s')
    def load_lib():
        log.info('Loading motif library')
        l_lib = []
        for mot in lib:
            # mod = run_x('open ' + mot[1])
            # vol = run_x('open ' + mot[2])
            mod = open_model(mot[1], log)
            vol = open_model(mot[2], log)
            l_lib.append([mot[0], mod, vol])
        return l_lib

    loaded_lib = load_lib()
    for k, pair in enumerate(pairs_list):
        if k > 0 and k % close_rate == 0:
            run_x('close session')
            loaded_lib = load_lib()

        #     p_id, mem = capture_memory_usage()
        #     with open(f'{p_id}.txt', 'a') as f:
        #         f.write('Closing session\n')
        #         f.write(f'{mem}\n')
        # p_id, mem = capture_memory_usage()
        # with open(f'{p_id}.txt', 'a') as f:
        #     f.write(f'{mem}\n')

        res_scores = score_residue(pair[0], pair[1], loaded_lib, log)

        res_name = os.path.basename(pair[1]).split('.')[0]
        name_lst = res_name.split('-')
        res_type = name_lst[0]
        res_nr = name_lst[1]
        chain = name_lst[2]
        residue_data = {keys.RES_TYPE: res_type,
                        keys.CHAIN: chain,
                        keys.RES_NR: int(res_nr),
                        keys.RES_MAP: os.path.basename(pair[0]),
                        keys.RES_MODEL: os.path.basename(pair[1]),
                        }
        for item in res_scores:
            # motif_name: correlation_m_matrix
            residue_data[item[0]] = f'{item[1]}_m_{item[2]}'

        lock.acquire()
        score_list.append(residue_data)
        if k % 10 == 0:
            scores_list_out = []
            for i in score_list:
                scores_list_out.append(i)
            scores_list_out.sort(key=operator.itemgetter(keys.CHAIN, keys.RES_NR))
            if os.path.exists(json_out_path):
                copy(json_out_path, json_out_path + '_bak')
            save_json(json_out_path, scores_list_out)
            if csv_out_path:
                with open(csv_out_path, mode='w') as csv_file:
                    fieldnames = [key for key in scores_list_out[0].keys()]
                    writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
                    writer.writeheader()
                    for item in scores_list_out:
                        writer.writerow(item)
        lock.release()


def score_structure(known_correlations, unknown_pairs, lib, json_out_path, csv_out_path=None, np=4):
    """
    Scores
    :param known_correlations:
    :param unknown_pairs:
    :param lib:
    :param json_out_path: out json file path
    :param csv_out_path: out csv file path
    :param np: number of cores
    :return:
    """
    log_tmp = 'process_logs'

    if not os.path.exists(log_tmp):
        os.mkdir(log_tmp)

    score_list = Manager().list()
    for entry in known_correlations:
        score_list.append(entry)

    split_pairs = split_list(unknown_pairs, np)
    thread_list = []
    lock = Lock()
    for batch_list in split_pairs:
        t = Process(target=slave, args=(batch_list, lib, score_list, lock, json_out_path, csv_out_path))
        thread_list.append(t)
    for thread in thread_list:
        thread.start()
    for thread in thread_list:
        thread.join()

    scores_list_out = []
    for i in score_list:
        scores_list_out.append(i)
    scores_list_out.sort(key=operator.itemgetter(keys.CHAIN, keys.RES_NR))
    copy(json_out_path, json_out_path + '_bak')
    save_json(json_out_path, scores_list_out)
    if csv_out_path:
        with open(csv_out_path, mode='w') as csv_file:
            fieldnames = [key for key in scores_list_out[0].keys()]
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            for item in scores_list_out:
                writer.writerow(item)
    # Combine logs
    combine_files(log_tmp, 'score_' + datetime.now().strftime("%Y-%m-%d") + '.log')
    if os.path.exists(log_tmp):
        rmtree(log_tmp)

def combine_files(dir_path, out_file):
    files = os.listdir(dir_path)
    files = [f for f in files if not f.startswith('.')]

    with open(out_file, 'w') as outfile:
        if len(files) > 0:
            for file in files:
                with open(os.path.join(dir_path, file), 'r') as infile:
                    for line in infile:
                        outfile.write(line)



def read_json(json_path):
    with open(json_path) as j:
        data = json.load(j)
    return data


def save_json(file_path, data):
    with open(file_path, 'w') as j:
        json.dump(data, j, indent=4)


def split_list(inp_list, nr):
    """
    Splits evenly a list
    :param inp_list: list
    :param nr: number of parts
    :return: list of "nr" lists
    """
    new_list = []
    nr_el = 1.0 / nr * len(inp_list)
    for i in range(nr):
        start = int(round(i * nr_el))
        end = int(round((i + 1) * nr_el))
        new_list.append(inp_list[start:end])
    return new_list


def check_known_correlations(inp_corr_list, lib):
    incomplete_indices = []
    for index, entry in enumerate(inp_corr_list):
        if len(entry.keys()) - 5 < len(lib) or "None" in entry.values():
            incomplete_indices.append(index)
        incomplete_indices.sort(reverse=True)

    for index in incomplete_indices:
        # del incomplete_indices[index]
        del inp_corr_list[index]
    return inp_corr_list


def exclude_known_pairs(pair_list, inp_corr_list):
    unknown_pairs = []
    for pair in pair_list:
        for entry in inp_corr_list:
            if os.path.basename(pair[0]) in entry.values():
                break
        else:
            unknown_pairs.append(pair)
    return unknown_pairs


def main():
    import time
    start = time.time()
    # print(sys.argv)
    st_arg = 1
    for item in sys.argv:
        if str(item).startswith('--'):
            st_arg += 1
    sys.argv = sys.argv[st_arg:]
    # print(sys.argv)
    parser = argparse.ArgumentParser(description='Score map')
    parser.add_argument("-p", "--pair_list", dest="pair_list", required=True,
                        help="List of residue map, residue pdb paths as json file")
    parser.add_argument("-l", "--lib", dest="lib", required=True, help="Motif library in_dir")
    parser.add_argument("-np", "--n_processors", dest="np", required=False, help="Number of processors")
    parser.add_argument("-r", "--recompute", dest="recompute", action='store_true',
                        help="Recalculate correlations")
    parser.add_argument("-o", "--out", dest="out", required=True, help="Out json file in_dir")
    args = parser.parse_args()
    # print(args.pair_list)
    pair_list = read_json(args.pair_list)
    lib = read_motif_lib(args.lib)
    if not args.recompute:
        inp_correlations = None
        try:
            inp_correlations = read_json(args.out)
        except (FileNotFoundError, json.decoder.JSONDecodeError) as _:
            try:
                inp_correlations = read_json(args.out + '_bak')
            except (FileNotFoundError, json.decoder.JSONDecodeError) as _:
                pass
        if inp_correlations:
            known_correlations = check_known_correlations(inp_correlations, lib)
            pair_list = exclude_known_pairs(pair_list, known_correlations)
            # for p in pair_list:
            #     print(p)
        else:
            known_correlations = []
    else:
        known_correlations = []

    head, tail = os.path.split(args.out)
    csv_path = os.path.join(head, tail.split('.')[0] + '.csv')
    score_structure(known_correlations, pair_list, lib, args.out,  csv_path, np=int(args.np))
    print(f'Run time = {time.time()-start}')
    run_x('exit')

main()




"""
bioUtils.py

A collection of functions for atomic models manipulations

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2019-10-09'


import os
import numpy as np
import string
import copy
from itertools import combinations
from Bio.PDB.Superimposer import Superimposer
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.MMCIFParser import MMCIFParser, FastMMCIFParser
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.mmcifio import MMCIFIO
from scipy.cluster.hierarchy import linkage, fcluster, maxdists
from scipy.spatial import distance as ssd
from . import nomenclature as nm
from threed_strudel.utils import modelMapUtils


def load_structure(model_path):
    """
    Creates Biopython structure object from file
    :param model_path: pdb, cif file in_dir
    :return: Biopython structure object
    """
    def copy_label_seq_id(in_path):
        parser = MMCIFParser(QUIET=True)
        parser._mmcif_dict = MMCIF2Dict(in_path)
        if "_atom_site.auth_seq_id" in parser._mmcif_dict:
            for i in parser._mmcif_dict["_atom_site.auth_seq_id"]:
                try:
                    int(i)
                except ValueError:
                    parser._mmcif_dict["_atom_site.auth_seq_id"] = parser._mmcif_dict["_atom_site.label_seq_id"]
        parser._build_structure(in_path.split('/')[-1].split('.')[0])
        return parser._structure_builder.get_structure()

    if model_path.split('.')[-1] == 'pdb' or model_path.split('.')[-1] == 'ent':
        parser = PDBParser(PERMISSIVE=1, QUIET=True)
        structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
    elif model_path.split('.')[-1] == 'cif':
        try:
            parser = MMCIFParser(QUIET=True)
            structure = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
        except ValueError:
            structure = copy_label_seq_id(model_path)
    else:
        raise Exception('Please provide the input residue in pdb or cif format')

    return structure


def load_structure_label_id(model_path):
    """
    Creates Biopython structure object from file
    :param model_path: pdb, cif file in_dir
    :return: Biopython structure object
    """
    # def copy_label_seq_id(in_path):
    parser = MMCIFParser(QUIET=True)
    parser._mmcif_dict = MMCIF2Dict(model_path)
    for field in ['_atom_site.auth_seq_id', '_atom_site.auth_comp_id', '_atom_site.auth_asym_id', '_atom_site.auth_atom_id']:
        parser._mmcif_dict[field] = copy.deepcopy(parser._mmcif_dict[field.replace('auth', 'label')])

    while True:
        for i, rec in enumerate(parser._mmcif_dict['_atom_site.auth_seq_id']):
            try:
                int(rec)
            except ValueError:
                for key, value in parser._mmcif_dict.items():
                    if key.startswith('_atom_site'):
                        del parser._mmcif_dict[key][i]
                break
        else:
            break
    parser._build_structure(model_path.split('/')[-1].split('.')[0])
    structure = parser._structure_builder.get_structure()

    return structure


def fast_load_model(model_path):
    """
    Creates Biopython structure object from file
    :param model_path: pdb, cif file in_dir
    :return: Biopython structure object
    """
    if model_path.split('.')[-1] == 'pdb' or model_path.split('.')[-1] == 'ent':
        parser = PDBParser(PERMISSIVE=1)
        model = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
    elif model_path.split('.')[-1] == 'cif':
        parser = FastMMCIFParser()
        model = parser.get_structure(model_path.split('/')[-1].split('.')[0], model_path)
    else:
        raise Exception('Please provide the input residue in pdb or cif format')
    return model


def save_model(model, out_path, preserve_atom_numbering=False):
    """
    Save residue to file
    :param model: structure object
    :param out_path: output file in_dir
    """
    ext = out_path.split('.')[-1]
    if ext == 'pdb' or ext == 'ent':
        io = PDBIO()
    else:
        io = MMCIFIO()
    io.set_structure(model)
    io.save(out_path, preserve_atom_numbering=preserve_atom_numbering)


def residues2structure(residues):
    """
    Builds a biopython structure object from a residue or a list of residue objects
    :param residues: residue object/ list of residue objects
    :return: structure object
    """

    if type(residues) not in [list, tuple]:
        residues = [residues]
    for res in residues:
        if res.level != 'R':
            raise Exception(f'Residue object expected (R), {res.level} given instead!')

    sb = StructureBuilder()
    sb.init_structure('structure')
    sb.init_seg(' ')
    try:
        res_model_id = set([r.parent.parent.id for r in residues])
    except AttributeError:
        res_model_id = {0}
    res_chain_id = set([r.parent.id for r in residues])
    for m in res_model_id:
        sb.init_model(m)
    for c in res_chain_id:
        sb.init_chain(c)
    for r in residues:
        model_id = r.parent.parent.id
        chain_id = r.parent.id
        sb.structure[model_id][chain_id].add(r.copy())
    for atom in sb.structure.get_atoms():
        atom.disordered_flag = 0

    return sb.structure


def rename_chains(structure):
    """
    Renames the chains using ascii uppercase lowercase and digits in order to minimise the chains names length
    :param structure: Biopython structure object
    :return: Biopython structure object
    """
    ch_ids = []
    for model in structure:
        for chain in model:
            if chain.id not in ch_ids:
                ch_ids.append(chain.id)

    symbols = string.ascii_uppercase + string.ascii_lowercase + string.digits
    base_set = [i for i in symbols]
    ids = base_set
    comb_nr = 2
    while len(ch_ids) > len(ids):
        comb = [''.join(i) for i in combinations(base_set, comb_nr)]
        ids = ids + comb
        comb_nr += 1

    translation = {}
    for i, c in enumerate(ch_ids):
        translation[c] = ids[i]

    for model in structure:
        for chain in model:
            chain.id = translation[chain.id]
    return structure


def get_cif_loop_item_pointer(lines, category, item):
    """
    Finds the position of a data item in a list cif file lines.
    Example: get_cif_loop_item_pointer(lines, '_atom_site', 'pdbx_PDB_model_num')
    :param lines: cif file as list of lines
    :param category: Category name
    :param item: Item name
    :return: Field number (starting from 0),
             Start data line
             End data line
    """
    field = None
    start_line = None
    end_line = None
    for n, line in enumerate(lines):
        if line.rstrip() == 'loop_':
            if lines[n + 1].split('.')[0] == category:
                k = 1
                while True:
                    key_words = lines[n + k].rstrip().split('.')
                    if key_words[0] == category:
                        if key_words[-1] == item:
                            field = k - 1
                        k += 1
                    else:
                        start_line = n + k
                        break
                break

    for n, line in enumerate(lines[start_line:]):
        if line.rstrip() == '#':
            end_line = start_line + n - 1
            break

    return field, start_line, end_line


def make_int_auth_seq_id(model_path, out_path=None):
    """
    Makes auth_seq_id integer numbers
    :param model_path:
    :param out_path:
    :return:
    """
    model_dict = MMCIF2Dict(model_path)
    auth_ids = model_dict['_atom_site.auth_seq_id']
    non_int_entities = {}
    first_id = ''
    entity = ''
    # Find sequences containing non integer auth_seq_id
    for i, a_id in enumerate(auth_ids):
        next_entity = model_dict['_atom_site.label_asym_id'][i]
        if next_entity != entity:
            first_id = a_id
            entity = next_entity
        try:
            int(a_id)

        except ValueError:

            if entity not in non_int_entities.keys():
                non_int_entities[entity] = first_id

    # Rename auth_seq_id. Try to use the first encountered auth_seq_id as starting number

    for _ in range(len(non_int_entities)):
        conversion = {}
        proc_entity = None
        new_id = 1
        for i, a_id in enumerate(auth_ids):
            entity = model_dict['_atom_site.label_asym_id'][i]
            if entity in non_int_entities.keys():
                if proc_entity is None:
                    proc_entity = entity
                    try:
                        new_id = int(non_int_entities[entity])
                    except ValueError:
                        new_id = 1
                    conversion[a_id] = new_id
            if entity == proc_entity:
                if a_id not in conversion.keys():
                    new_id += 1
                    conversion[a_id] = new_id
                auth_ids[i] = str(conversion[a_id])
        del non_int_entities[proc_entity]

    io = MMCIFIO()
    io.set_dict(model_dict)
    if out_path is not None:
        io.save(out_path)
    else:
        io.save(model_path)


def classify_chains(structure, delta=0.1):
    """
    Classifies structure chains based on sequence_dir identity and rms
    :param delta: rms distance between classes
    :param structure: Biopython structure object
    :return: list of chain classes
    """
    # Find pairs of identical chains

    sup = Superimposer()
    edges = []
    if structure.get_level() == 'S':
        model = structure[0]
    elif structure.get_level() == 'M':
        model = structure
    else:
        return None, None

    chain_ids = [chain.get_id() for chain in model]
    if len(chain_ids) < 2:
        return [chain_ids], None
    distance_matrix = np.zeros(shape=(len(chain_ids), len(chain_ids)))

    for chains in combinations(model, 2):
        different = False
        # Exclude hetero residues
        residues1 = chains[0].get_list()
        residues1 = [res for res in residues1 if res.get_id()[0] == " "]
        residues2 = chains[1].get_list()
        residues2 = [res for res in residues2 if res.get_id()[0] == " "]
        if len(residues1) == len(residues2) and len(residues1) != 0:
            for i in range(len(residues1)):
                if residues1[i].get_resname() != residues2[i].get_resname():
                    different = True
                    break
            else:
                # Superimpose chains with identical sequence_dir
                moving = []
                fixed = []
                for res in residues1:
                    for atom in res.get_atoms():
                        moving.append(atom)
                for res in residues2:
                    for atom in res.get_atoms():
                        fixed.append(atom)
                if len(fixed) == len(moving):
                    sup.set_atoms(fixed, moving)
                    edges.append((chains[0].get_id(), chains[1].get_id(), sup.rms))
                    distance_matrix[chain_ids.index(chains[0].get_id()), chain_ids.index(chains[1].get_id())] = sup.rms
                    distance_matrix[chain_ids.index(chains[1].get_id()), chain_ids.index(chains[0].get_id())] = sup.rms
        else:
            different = True

        if different:
            distance_matrix[chain_ids.index(chains[0].get_id()), chain_ids.index(chains[1].get_id())] = delta * 2
            distance_matrix[chain_ids.index(chains[1].get_id()), chain_ids.index(chains[0].get_id())] = delta * 2

    # Cluster using the distance matrix
    # print(distance_matrix)
    Z = linkage(ssd.squareform(distance_matrix), method="complete")
    max_dist = maxdists(Z)
    clusters_ind = fcluster(Z, delta, criterion='distance')

    clusters = []
    for i in range(max(clusters_ind)):
        clusters.append([])
    for ind, c in enumerate(clusters_ind):
        clusters[c - 1].append(chain_ids[ind])



    sorted_clusters = []
    for chain in chain_ids:
        for cluster in clusters:
            if chain in cluster and cluster not in sorted_clusters:
                tmp_cluster = []
                for c in chain_ids:
                    if c in cluster:
                        tmp_cluster.append(c)
                sorted_clusters.append(tmp_cluster)

    return sorted_clusters, list(max_dist[-len(sorted_clusters)+1:])


def classify_chains_(structure, delta=0.1):
    """
    Classifies structure chains based on sequence_dir identity and rms
    :param delta: rms distance between classes
    :param structure: Biopython structure object
    :return: list of chain classes
    """
    # Find pairs of identical chains

    sup = Superimposer()
    # edges = []
    if structure.get_level() == 'S':
        model = structure[0]
    elif structure.get_level() == 'M':
        model = structure
    else:
        return None, None

    chain_ids = [chain.get_id() for chain in model]
    if len(chain_ids) < 2:
        return [chain_ids], None
    distance_matrix = np.zeros(shape=(len(chain_ids), len(chain_ids)))

    twin_classes = []
    has_twins = []

    for chains in combinations(model, 2):
        ch_id1 = chains[0].get_id()
        ch_id2 = chains[1].get_id()
        different = False
        # Exclude hetero residues
        residues1 = chains[0].get_list()
        residues1 = [res for res in residues1 if res.get_id()[0] == " "]
        residues2 = chains[1].get_list()
        residues2 = [res for res in residues2 if res.get_id()[0] == " "]
        if len(residues1) == len(residues2) and len(residues1) != 0:
            for i in range(len(residues1)):
                if residues1[i].get_resname() != residues2[i].get_resname():
                    different = True
                    break
            else:
                duplex = [ch_id1, ch_id2]
                print(duplex)
                print(has_twins)
                if ch_id1 not in has_twins and ch_id2 not in has_twins:
                    twin_classes.append(duplex)
                    has_twins.append(ch_id1)
                    has_twins.append(ch_id2)
                    print(has_twins)
                else:
                    for i, group in enumerate(twin_classes):
                        if ch_id1 in group or ch_id2 in group:
                            group.append(ch_id1)
                            group.append(ch_id2)
                print(twin_classes)

    print(twin_classes)
    for i, group in enumerate(twin_classes):
        twin_classes[i] = list(set(group))
    print(twin_classes)


    # for chains in combinations(residue, 2):
    #     different = False
    #     # Exclude hetero residues
    #     residues1 = chains[0].get_list()
    #     residues1 = [res for res in residues1 if res.get_id()[0] == " "]
    #     residues2 = chains[1].get_list()
    #     residues2 = [res for res in residues2 if res.get_id()[0] == " "]
    #     if len(residues1) == len(residues2) and len(residues1) != 0:
    #         for i in range(len(residues1)):
    #             if residues1[i].get_resname() != residues2[i].get_resname():
    #                 different = True
    #                 break
    #         else:
    #             # Superimpose chains with identical sequence_dir
    #             moving = []
    #             fixed = []
    #             for res in residues1:
    #                 for atom in res.get_atoms():
    #                     moving.append(atom)
    #             for res in residues2:
    #                 for atom in res.get_atoms():
    #                     fixed.append(atom)
    #             if len(fixed) == len(moving):
    #                 sup.set_atoms(fixed, moving)
    #                 # edges.append((chains[0].get_id(), chains[1].get_id(), sup.rms))
    #                 distance_matrix[chain_ids.index(chains[0].get_id()), chain_ids.index(chains[1].get_id())] = sup.rms
    #                 distance_matrix[chain_ids.index(chains[1].get_id()), chain_ids.index(chains[0].get_id())] = sup.rms
    #     else:
    #         different = True
    #
    #     if different:
    #         distance_matrix[chain_ids.index(chains[0].get_id()), chain_ids.index(chains[1].get_id())] = delta * 2
    #         distance_matrix[chain_ids.index(chains[1].get_id()), chain_ids.index(chains[0].get_id())] = delta * 2
    #
    # # Cluster using the distance matrix
    # # print(distance_matrix)
    # Z = linkage(ssd.squareform(distance_matrix), method="complete")
    # max_dist = maxdists(Z)
    # clusters_ind = fcluster(Z, delta, criterion='distance')
    #
    # clusters = []
    # for i in range(max(clusters_ind)):
    #     clusters.append([])
    # for ind, c in enumerate(clusters_ind):
    #     clusters[c - 1].append(chain_ids[ind])
    #
    #
    #
    # sorted_clusters = []
    # for chain in chain_ids:
    #     for cluster in clusters:
    #         if chain in cluster and cluster not in sorted_clusters:
    #             tmp_cluster = []
    #             for c in chain_ids:
    #                 if c in cluster:
    #                     tmp_cluster.append(c)
    #             sorted_clusters.append(tmp_cluster)
    #
    # return sorted_clusters, list(max_dist[-len(sorted_clusters)+1:])

def select_best_chains(structure, map_object, chain_classes, threshold=None):
    """
    Selects best fitted chain in each class
    :param structure: Biopython structure object
    :param map_object: map object created by mapParser
    :param chain_classes: a list of chain classes
    :param threshold: map threshold level
    :return: a list containing best fitted chain from each class
    """

    if threshold is None:
        threshold = modelMapUtils.find_threshold(structure, map_object, inclusion_fraction=90)

    if structure.get_level() == 'S':
        model = structure[0]
    elif structure.get_level() == 'M':
        model = structure
    else:
        return None
    filtered = []
    for _class in chain_classes:
        if len(_class) == 1:
            filtered.append(_class[0])
        else:
            best_chain = None
            best_inclusion = -1
            for chain in _class:
                inclusion = modelMapUtils.atom_inclusion(model[chain], map_object, threshold)[2]
                if inclusion > best_inclusion:
                    best_chain = chain
                    best_inclusion = inclusion
            filtered.append(best_chain)

    return filtered


def mmcif2pdb_biopython(cif_file, pdb_file=None):
    """
    Converts MMCIF int PDB file when possible
    :param cif_file: MMCIF file
    :param pdb_file: PDB file
    :return: PDB file in_dir
    """
    if pdb_file is None:
        head, tail = os.path.split(cif_file)
        pdb_file = os.path.join(head, tail.split('.')[0] + '.pdb')
    s = load_structure(cif_file)
    l = len([a for a in s.get_atoms()])
    if l > 99999:
        return None
    try:
        save_model(s, pdb_file)
        return pdb_file
    except TypeError:
        pass


def pdb2mmcif(pdb_file, cif_file=None):
    """
    Converts PDB to MMCIF file
    :param cif_file: MMCIF file
    :param pdb_file: PDB file
    :return: MMCIF file in_dir
    """
    if cif_file is None:
        head, tail = os.path.split(pdb_file)
        cif_file = os.path.join(head, tail.split('.')[0] + '.cif')
    s = load_structure(pdb_file)
    save_model(s, cif_file)

    return cif_file


def shift_coord(trans_matrix, structure):
    """
    Apply translation matrix to residue
    :param trans_matrix: translation matrix
    :param structure: biopython structure object
    """
    def shift():
        atom.get_coord()[0] = atom.get_coord()[0] - trans_matrix[0]
        atom.get_coord()[1] = atom.get_coord()[1] - trans_matrix[1]
        atom.get_coord()[2] = atom.get_coord()[2] - trans_matrix[2]
    try:
        for residue in structure.get_residues():
            for atom in residue:
                try:
                    disordered = atom.disordered_get_id_list()
                except AttributeError:
                    shift()
                else:
                    for id_ in disordered:
                        atom.disordered_select(id_)
                        shift()
    except AttributeError:
        for atom in structure:
            try:
                disordered = atom.disordered_get_id_list()
            except AttributeError:
                shift()
            else:
                for id_ in disordered:
                    atom.disordered_select(id_)
                    shift()

def del_main_chain(residue):
    """
    Delete main chain atoms
    :param residue: BIO.PDB object
    :return: BIO.PDB object
    """
    for atom in residue:
        if atom.get_name() == "O":
            del residue[atom.get_name()]
    for atom in residue:
        if atom.get_name() == "N":
            del residue[atom.get_name()]
    for atom in residue:
        if atom.get_name() == "C":
            del residue[atom.get_name()]
    return residue

def calc_residue_static_pairwise_rmsd(model1, model2):
    """
    Calculate pairwise RMSD for residue structure objects without superimposing
    Takes into account residues symmetry
    :param model1: structure object
    :param model2: structure object
    :return: RMSD value
    """
    atoms1 = [atom for atom in model1.get_atoms() if not atom.get_name().upper().startswith('H')]
    atoms2 = [atom for atom in model2.get_atoms() if not atom.get_name().upper().startswith('H')]
    names1 = [atom.get_name() for atom in atoms1]
    names2 = [atom.get_name() for atom in atoms2]

    if len(atoms1) > len(atoms2):
        atoms1 = [atom for atom in atoms1 if atom.get_name() in names2]
    else:
        atoms2 = [atom for atom in atoms2 if atom.get_name() in names1]
    atoms1.sort(key=lambda x: x.get_name())
    atoms2.sort(key=lambda x: x.get_name())

    array1, array2 = np.array([atoms1[0].get_coord()]), np.array([atoms2[0].get_coord()])
    for i in range(1, len(atoms1)):
        array1 = np.concatenate((array1, np.array([atoms1[i].get_coord()])), axis=0)
        array2 = np.concatenate((array2, np.array([atoms2[i].get_coord()])), axis=0)
    rmsd_ = rmsd(array1, array2)

    res_type = model1.get_resname().lower()
    all_res_symm_pairs = nm.SYMMETRIC_ATOM_PAIRS
    if res_type not in all_res_symm_pairs.keys():
        return rmsd_

    symmetric_pairs = all_res_symm_pairs[res_type]
    swap_indices = [[],[]]
    for i in range(0, len(symmetric_pairs), 2):
        for k, atom in enumerate(atoms2):
            if atom.get_name().lower() == symmetric_pairs[i]:
                swap_indices[0].append(k)
            elif atom.get_name().lower() == symmetric_pairs[i+1]:
                swap_indices[1].append(k)
    for i in range(len(swap_indices[0])):
        atoms2[swap_indices[0][i]], atoms2[swap_indices[1][i]] = \
            atoms2[swap_indices[1][i]], atoms2[swap_indices[0][i]]

    array3 = np.array([atoms2[0].get_coord()])
    for i in range(1, len(atoms2)):
        array3 = np.concatenate((array3, np.array([atoms2[i].get_coord()])), axis=0)
    rmsd2 = rmsd(array1, array3)
    return min(rmsd_, rmsd2)


def rmsd(array1, array2):
    """
    Calculates RMSD between two sets of coordinates without alignment
    :param array1: numpy 2d array
    :param array2: numpy 2d array
    :return: rmsd
    """
    if np.shape(array1) != np.shape(array1):
        raise Exception('The coordinate arrays must have the same dimensions')
    dif = array1 - array2
    return np.sqrt(np.mean(np.sum(dif * dif, axis=1)))


def superimpose_n_ca_c(str_fixed, str_moving):
    """
    Superimpose 2 models containing single residues using N, CA, C atom selection
    :param str_fixed: BIO.PDB structure object
    :param str_moving: BIO.PDB structure object
    """
    fixed = []
    moving = []
    for atom in [a for a in str_fixed.get_atoms()]:
        if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
            fixed.append(atom)
    for atom in [a for a in str_moving.get_atoms()]:
        if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
            moving.append(atom)
    sup = Superimposer()
    sup.set_atoms(fixed, moving)
    sup.apply([a for a in str_moving.get_atoms()])
    return str_moving



import json
import csv
import os
import matplotlib.pyplot as plt
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter,
                               AutoMinorLocator)

from matplotlib.pyplot import figure, close
import numpy as np

def read_json(json_path):
    with open(json_path) as j:
        data = json.load(j)
    return data


def save_json(file_path, data):
    with open(file_path, 'w') as j:
        json.dump(data, j, indent=4)


def get_percentage(file, top_n):
    data = read_json(file)
    if len(data) == 0:
        return 0, 0
    nr_1 = 0
    nr_top = 0
    for entry in data:
        if entry['place'] == 0:
            nr_1 += 1
        if -1 < entry['place'] < top_n:
            nr_top += 1
    cor_perc = nr_1 / len(data) * 100
    top_n_perc = nr_top / len(data) * 100
    print(len(data))
    return cor_perc, top_n_perc


def single_bar(x, y, out=None):
    fig, ax = plt.subplots()
    plt.bar(x, y)
    plt.ylim(0, 100)
    plt.xticks([i for i in range(nr[0], nr[-1] + 1, 5)])
    ax.yaxis.set_minor_locator(MultipleLocator(5))
    ax.set_xlabel('Fragment length')
    ax.set_ylabel('% Correct predictions')
    if out is not None:
        fig.savefig(out)
    else:
        plt.show()

def double_bar(title, x1, x2, y1, y2, out=None):
    bar_width = 0.35
    fig, ax = plt.subplots()
    print(dir(fig))
    fig.set_size_inches(10,4)
    ax.bar(x1, y1, bar_width, color='b', label='Buccaneer')
    ax.bar(np.array(x2) + bar_width, y2, bar_width, color='g', label='Strudel')
    ax.set_xlabel('Fragment length')
    ax.set_ylabel('% Correct predictions')
    ax.yaxis.set_minor_locator(MultipleLocator(5))
    ax.legend()
    plt.title(title)
    if out is not None:
        fig.savefig(out)
    else:
        plt.show()


def get_data(path):
    top_number = 10
    files = os.listdir(path)
    j_files = [f for f in files if f.endswith('.json') and not f.startswith('.')]
    try:
        j_files.sort(key=lambda z: int(z.split('_')[0]))
    except ValueError:
        j_files.sort(key=lambda z: int(z.split('.')[0]))
    nr = []
    perc = []
    first_10_perc = []
    for f in j_files:
        try:
            n = int(f.split('_')[0])
        except ValueError:
            n = int(f.split('.')[0])
        nr.append(n)
        j_path = os.path.join(path, f)
        cor_perc, top_n = get_percentage(j_path, top_number)
        perc.append(cor_perc)
        first_10_perc.append(top_n)
    return nr, perc

# j_dir = '/Volumes/data/Work/test_segm/out_6/vs_2.5-3_formated/res_jsons_1000_60'
#j_dir = '/Volumes/data/Work/test_segm/out_6/vs_2.5-3_formated/buc_jsons'
# j_dir = '/Volumes/data/Work/test_segm/out_6/vs_2.5-3_formated/jsons_blind'
# j_dir = '/Volumes/data/bucc_find_seq/results/20584'
# j_dir = '/Volumeses/data/find_seq/0711_strudel/jsons_75_74-'
# j_dir = '/Volumes/data/find_seq/untitled_folder/20584out_sd1/jsons_75'
# top_number = 10
# files = os.listdir(j_dir)
# j_files = [f for f in files if f.endswith('.json') and not f.startswith('.')]
# j_files.sort(key=lambda z: int(z.split('_')[0]))
#
# nr = []
# perc = []
# first_10_perc = []
# for f in j_files:
#     n = int(f.split('_')[0])
#     nr.append(n)
#     j_path = os.path.join(j_dir, f)
#     cor_perc, top_n = get_percentage(j_path, top_number)
#     perc.append(cor_perc)
#     first_10_perc.append(top_n)
# #
# print(nr)
# # fig = figure(num=1, figsize=(4, 4), dpi=200, facecolor='w', edgecolor='k')
# fig, ax = plt.subplots()
# plt.bar(nr, perc)
# plt.ylim(0, 100)
# plt.xticks([i for i in range(nr[0], nr[-1] + 1, 5)])
# ax.yaxis.set_minor_locator(MultipleLocator(5))

# fig.savefig("/Volumes/data/find_seq/plots/" + '20584_strudel_sd2.png')

# strud = '/Users/andrei/sequence_data/6272_strudel/res_jsons_1000'
# buc = '/Users/andrei/sequence_data/6272_buc/jsons'
# out = '/Users/andrei/sequence_data/plots_double/6272_dep.png'
#
# strud = '/Users/andrei/sequence_data/6272_strud_built_results/sd2/jsons_75'
# buc = '/Users/andrei/sequence_data/6272_buc_bui/jsons'
# out = '/Users/andrei/sequence_data/plots_double/6272_built_sd2_75.png'

# strud = '/Users/andrei/sequence_data/6272_strud_built_results/sd1/jsons_75'
# buc = '/Users/andrei/sequence_data/6272_buc_bui/jsons'
# out = '/Users/andrei/sequence_data/plots_double/6272_built_sd1_75.png'

# strud = '/Users/andrei/sequence_data/20584_strud_bui/sd1/jsons_75'
# buc = '/Users/andrei/sequence_data/20584_buc_bui/jsons'
# out = '/Users/andrei/sequence_data/plots_double/20584_built_sd1_75.png'

# strud = '/Users/andrei/sequence_data/20584_strud_bui/sd1/jsons_60'
# buc = '/Users/andrei/sequence_data/20584_buc_bui/jsons'
# out = '/Users/andrei/sequence_data/plots_double/20584_built_sd1_60.png'

strud = '/Volumes/data/find_seq/0703/strudel/0703/jsons_75'
buc = '/Volumes/data/find_seq/0703/buc/jsons'
out = '/Users/andrei/sequence_data/plots_double/0703_dep_sd1_75.png'


# st_nr, st_perc = get_data(strud)
# buc_nr, buc_perc = get_data(buc)
# print(st_nr)
# print(buc_nr)

# double_bar(buc_nr, st_nr, buc_perc, st_perc, out)


def all_plots(in_path, out_path):
    if not os.path.exists(out_path):
        os.makedirs(out_path)

    entries = os.listdir(in_path)
    entries = [p for p in entries if p!= 'plots']
    for entry in entries:
        code, resol = entry.split('_')
        buc_path = os.path.join(in_path, entry, 'buc/jsons')
        strud_fold = os.path.join(in_path, entry, 'strudel')
        percs = os.listdir(strud_fold)
        for fldr in percs:
            aa = r"$\mathring{A}$"
            strud_path = os.path.join(strud_fold, fldr)
            title = f'Entry: {code}, resolution: {resol} {aa}'
            print(title)
            perc = fldr.split('_')[-1]
            st_nr, st_perc = get_data(strud_path)
            buc_nr, buc_perc = get_data(buc_path)
            out = os.path.join(out_path, f'{code}_{resol}_{perc}.png')
            double_bar(title, buc_nr, st_nr, buc_perc, st_perc, out)

all_in = '/Volumes/data/find_seq/all_results'
all_out = '/Volumes/data/find_seq/all_results/plots'

all_plots(all_in, all_out)


"""
rotamerClassifier.py

Splits residues of the same type into specified number of rotameric classes
based on RMSD.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Superimposer import Superimposer
from random import randint
import os
import copy


class RotamerClassifier:
    """
    Class for rotamer classification
    """
    class Atoms:
        res = {'phe': ['CB', 'CG', 'CD1', 'CD2'],
               'val': ['CB', 'CG1', 'CG2']

               }



    def __init__(self, path, rotamer_number):
        self.path = path
        self.map_file_format = 'map'
        self.str_object_list = []
        self.rotamer_number = rotamer_number
        self.create_pdb_list()


    def create_pdb_list(self):
        """
        Create a list of dictionaries with pdb file name as keys

        """
        files = os.listdir(self.path)
        pdb_names = [i for i in files if i.endswith('.pdb')]
        parser = PDBParser(PERMISSIVE=1)
        for element in pdb_names:
            name = element.split('.')[0]
            new_str = parser.get_structure(name, self.path + '/' + element)
            valid_str = True
            atoms = self.Atoms
            for residue in new_str.get_residues():
                type = residue.get_resname().lower()

            for atom in atoms.res[type]:
                if atom not in [a.get_id() for a in new_str.get_atoms()]:
                    valid_str = False
            if valid_str:
                self.str_object_list.append(new_str)
            else:
                print(new_str.id)

    def superimpose_n_ca_c(self, str_fixed, str_moving):
        """
        Superimpose 2 models containing single residues using N, CA, C atom selection
        """
        fixed = []
        moving = []
        for atom in [a for a in str_fixed.get_atoms()]:
            if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
                fixed.append(atom)
        for atom in [a for a in str_moving.get_atoms()]:
            if atom.get_id() == 'CA' or atom.get_id() == 'C' or atom.get_id() == 'N':
                moving.append(atom)
        sup = Superimposer()
        sup.set_atoms(fixed, moving)
        sup.apply([a for a in str_moving.get_atoms()])
        return str_moving

    def superimpose_all(self):
        """
        Superimpose all structures to the first one in the list
        """
        i = 1
        while i < len(self.str_object_list):
            self.str_object_list[i] = self.superimpose_n_ca_c(self.str_object_list[0], self.str_object_list[i])
            i += 1

    def calc_pairwise_rmsd(self, model1, model2):
        """
        Calculate pairwise RMSD for structure objects
        :param model1: structure object
        :param model2: structure object
        :return: RMSD value
        """
        summ = 0
        i = 0
        for atom1 in [a for a in model1.get_atoms()]:
            for atom2 in [a for a in model2.get_atoms()]:
                if atom1.get_id() == atom2.get_id():
                    d = atom1 - atom2
                    summ += d * d
                    i += 1
        rmsd = (summ / i) ** (0.5)
        return rmsd

    def save_pdb(self, model, out_path):
        """
        Save residue as pdb
        :param model: BIO.PDB structure object
        :param out_path: out file in_dir
        """
        io = PDBIO()
        io.set_structure(model)
        io.save(out_path)

    def concatenate_pdb(self, model_dir, out_file):
        """
        Merge all pdb files in a directory
        """
        files = os.listdir(model_dir)
        pdb_files = [i for i in files if i.endswith('.pdb')]
        nr = 0
        lines = []
        with open(model_dir + '/' + out_file, 'w') as reference:
            for model in pdb_files:
                nr += 1
                with open(model_dir + '/' + model, 'r') as model_text:
                    lines.append('MODEL     ' + str(nr) + '\n')
                    for line in model_text:
                        if line != 'END\n':
                            lines.append(line)
                        else:
                            lines.append('ENDMDL\n')
            for line in lines:
                reference.write(line)
            reference.write('END\n')

    def split_rotamers(self):
        """
        Group the models into classes based on pairwise RMSD
        :return: a list of classes
        """
        # Find the initial reference models for each class
        classes = []
        for group in range(self.rotamer_number):
            classes.append([])
        classes[0].append(self.str_object_list[randint(0, len(self.str_object_list))])
        class_index = 1
        while class_index < self.rotamer_number:
            rmsd_list = []

            for structure in self.str_object_list:
                rmsd = 0
                for group in classes:
                    if group != []:
                        rmsd += self.calc_pairwise_rmsd(group[0], structure)
                rmsd_list.append(rmsd)
            max_rmsd = max(rmsd_list)
            classes[class_index].append(self.str_object_list[rmsd_list.index(max_rmsd)])
            class_index += 1
        # Add the remaining models to the corresponding classes
        i = 0
        while i < len(self.str_object_list):
            class_index = 0
            class_rmsd =[]
            while class_index < len(classes):
                rmsd = self.calc_pairwise_rmsd(classes[class_index][0], self.str_object_list[i])
                class_rmsd.append(rmsd)
                class_index += 1
            if 0.0 not in class_rmsd:
                min_rms = min(class_rmsd)
                classes[class_rmsd.index(min_rms)].append(self.str_object_list[i])
            i += 1
        # Find representative models within classes and repeat classification using
        # representative models as references. Execute while representative models change
        new_repr_set = []
        #
        while True:
            # Find representative models within classes
            class_index = 0
            while class_index < len(classes):
                rmsd_list = []
                length = len(classes[class_index])
                k = 0
                while k < length:
                    i = 0
                    rmsd = 0
                    while i < length:
                        rmsd += self.calc_pairwise_rmsd(classes[class_index][k], classes[class_index][i])
                        i += 1
                    rmsd_list.append(rmsd)
                    k += 1
                min_rms = min(rmsd_list)
                representative_model = classes[class_index][rmsd_list.index(min_rms)]
                classes[class_index] = []
                classes[class_index].append(representative_model)
                class_index += 1
            old_repr_set = copy.deepcopy(new_repr_set)
            new_repr_set = [x[0].id for x in classes]
            print(new_repr_set)
            # Repeat classification using representative models as references
            i = 0
            while i < len(self.str_object_list):
                class_index = 0
                class_rmsd =[]
                while class_index < len(classes):
                    rmsd = self.calc_pairwise_rmsd(classes[class_index][0], self.str_object_list[i])
                    class_rmsd.append(rmsd)
                    class_index += 1
                if 0.0 not in class_rmsd:
                    min_rms = min(class_rmsd)
                    classes[class_rmsd.index(min_rms)].append(self.str_object_list[i])
                i += 1
            if old_repr_set == new_repr_set:
                for cl in classes:
                    tmp_name = cl[0].id + '_representative'
                    cl[0].id = tmp_name
                break
        return classes


def main():
    pdb_dir = 'test_new/models_and_maps'
#    pdb_dir = 'x-ray/5o4i/tyr/residue'
    rotamers = 3
    cl = RotamerClassifier(pdb_dir, rotamers)
    cl.superimpose_all()
    classes = cl.split_rotamers()

    for i in range(rotamers):
        path = pdb_dir + '/' + 'rot_' + str(i + 1)
        if not os.path.exists(path):
            os.mkdir(path)
        classes[i].sort(key=lambda x: (int(x.id.split('-')[1]), x.id.split('-')[2]))
        for k in classes[i]:
            cl.save_pdb(k, path + '/' + k.id + '.pdb')
        cl.concatenate_pdb(path, 'class_' + str(i + 1) + '.pdb')


if __name__ == '__main__':
    main()





import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
     name='threed_strudel',
     version='0.5',
     author="Andrei Istrate",
     author_email="andrei@ebi.ac.uk",
     description="Strudel package",
     include_package_data=True,
     scripts=['threed_strudel/bin/strudel_mapMotifValidation.py', 'threed_strudel/bin/strudel_mapAveraging.py',
              'threed_strudel/bin/strudel_penultimateClassifier.py', 'threed_strudel/bin/strudel_chopModelMapMPI.py',
              'threed_strudel/bin/strudel_setChimeraX.py'],
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "Operating System :: OS Independent"],
     install_requires=['scipy', 'biopython', 'numpy', 'mrcfile', 'psutil'],
     python_requires='>=3.6',
     license="Apache License"
 )


"""
mapAveraging.py

Superimpose and average maps using atom models as guide.
This code use maps in mrc format
This code uses Chimera for superimposing. UCSF Chimera packages are
imported using the pychimera module.

Copyright [2013] EMBL - European Bioinformatics Institute
Licensed under the Apache License, Version 2.0 (the
"License"); you may not use this file except in
compliance with the License. You may obtain a copy of
the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied. See the License for the
specific language governing permissions and limitations
under the License.
"""

__author__ = 'Andrei Istrate'
__email__ = 'andrei@ebi.ac.uk'
__date__ = '2018-05-29'

import sys
import os
import time
from shutil import copy2
import argparse
from datetime import datetime
import numpy as np
import json
from collections import Counter
import subprocess
import threed_strudel.configure as config
import threed_strudel.utils.functions as func
from threed_strudel.parse.mapParser import MapParser
import threed_strudel.average


class MapAveraging:
    """
    Class for map superimposing using atom models as a guide and map averaging
    """

    def __init__(self, model_dir,
                 rotamer_classes_dir,
                 rot_prefix='rot_',
                 map_class_prefix='class_',
                 box_size=14,
                 motifs_dir=None,
                 scaling='standardise',
                 warning_level='info',
                 log_path=None,
                 motif_suffix='',
                 min_reliable_nr=10):

        self.scaling = scaling
        self.motif_suffix = motif_suffix
        self.log = self.setup_log(log_path, warning_level)
        self.chimera_path = config.CHIMERA_PATH
        basedir = os.path.dirname(os.path.abspath(__file__))
        self.superimpose_chimera_script = os.path.join(basedir, 'superimpose_chimerax.py')
        self.final_box_size = box_size
        self.global_reference_path = self.get_global_reference_path()
        self.rotamer_classes_dir = os.path.abspath(rotamer_classes_dir)
        self.models_dir = os.path.abspath(model_dir)
        self.motifs_dir = motifs_dir
        self.warning_level = warning_level
        self.rot_prefix = rot_prefix
        self.map_class_prefix = map_class_prefix
        self.statistics = []
        self.rotamers = []
        self.classes = []
        self.representative_structures = []
        self.set_env()
        self.map_suffix = None
        self.representative = ''
        self.min_reliable_number = min_reliable_nr
        self.unreliable_motif_folder = 'unreliable'
        self.unreliable_motif_dir = os.path.join(self.motifs_dir, 'unreliable')
        if not os.path.exists(self.unreliable_motif_dir):
            os.makedirs(self.unreliable_motif_dir)
        self.run_main()

    @staticmethod
    def setup_log(log_path, warning_level):
        aver_log = func.setup_logger('average_log', log_path, warning_level=warning_level)
        return aver_log

    @staticmethod
    def get_global_reference_path():
        head = os.path.dirname(threed_strudel.average.__file__)
        path = os.path.join(head, 'inp/global_reference.pdb')
        return path

    def run_main(self):
        start = time.time()

        if len(self.classes) > 0:
            for index, class_name in enumerate(self.classes):
                class_dict = self.get_class_info(self.rotamers[index], class_name)

                # Do the averaging using the atomic models as reference
                out_folder = 'model_super'
                out_dir = os.path.join(self.rotamer_classes_dir, class_name, out_folder)
                class_dict['out_dir'] = out_dir
                self.create_dir(out_dir)

                if len(class_dict['class_pairs']) > 0:
                    t = time.time()
                    self.log.info('Superimposing maps using residue as a guide...')
                    self.superimpose(class_dict, fit_map=False)
                    self.log.debug('spent on superimposing %s', func.report_elapsed(t))
                    t = time.time()
                    self.log.info("Averaging the residue superimposed maps")
                    aver_map = self.average(class_dict)
                    self.log.debug('spent on averaging %s', func.report_elapsed(t))
                    # Copy the averaged densities to the motifs directory
                    name = os.path.basename(class_dict['reference_model'])
                    aligned_reference = os.path.join(class_dict['out_dir'], name)
                    copy2(aligned_reference, os.path.join(self.rotamer_classes_dir, class_name))
                    # Additional map to average map fitting ang repeat averaging
                    out_folder = 'map_super'
                    out_dir = os.path.join(self.rotamer_classes_dir, class_name, out_folder)
                    class_dict['out_dir'] = out_dir
                    self.create_dir(out_dir)
                    if aver_map is not None:
                        class_dict['reference_model'] = aligned_reference
                        class_dict['reference_mrc'] = aver_map
                        t = time.time()
                        self.log.info('Superimposing maps using residue as a guide followed by map to map fit...')
                        self.superimpose(class_dict, fit_map=True)
                        self.log.debug('spent on superimposing %s', func.report_elapsed(t))
                        t = time.time()
                        self.log.info("Averaging the map to map superimposed maps")
                        aver_map = self.average(class_dict)
                        self.log.debug('spent on averaging %s', func.report_elapsed(t))

                        # Copy the reference atomic models to the motifs directory
                        prefix = name.split('-')[0] + '_rotamer_' + class_name.split('_')[-1] + self.motif_suffix
                        map_motif_name = prefix + self.motif_suffix + '.mrc'
                        model_motif_name = prefix + self.motif_suffix + '.cif'

                        if len(class_dict['class_pairs']) >= self.min_reliable_number:
                            dst = self.motifs_dir
                        else:
                            dst = self.unreliable_motif_dir
                        copy2(aligned_reference, os.path.join(dst, model_motif_name))
                        copy2(aver_map, os.path.join(dst, map_motif_name))

        text = func.report_elapsed(start)
        self.log.info('\nElapsed: {}\n{:_^100}'.format(text, ''))

    @staticmethod
    def create_dir(path):
        """
        Create new directory
        :param path: new directory in_dir
        """
        if not os.path.exists(path):
            os.mkdir(path)

    def set_env(self):
        """
        For each folder named rot_prefix* create a folder named map_class_prefix*
        """
        date_time = datetime.now().strftime("%H:%M:%S %Y-%m-%d")
        text = '{:_^100}'.format('MapAveraging') + '\n\nStarted: {}'.format(date_time)
        self.log.info('%s\n\n', text)
        self.log.info("Searching rotamers in %s directory", self.rotamer_classes_dir)
        rot_folder_lst = [item for item in os.listdir(self.rotamer_classes_dir) if item.startswith(self.rot_prefix)]
        rot_folder_lst.sort()

        for rot_folder in rot_folder_lst:
            self.rotamers.append(rot_folder)
            suffix = rot_folder.split(self.rot_prefix)[-1]
            map_class_folder = self.map_class_prefix + suffix
            self.classes.append(map_class_folder)
            class_dir = os.path.join(self.rotamer_classes_dir, map_class_folder)
            self.create_dir(class_dir)

        if self.motifs_dir is None:
            self.motifs_dir = os.path.join(self.rotamer_classes_dir, 'motifs')
        if not os.path.exists(self.motifs_dir):
            os.makedirs(self.motifs_dir)

        self.log.info("The following rotamer folders were found:")
        for r in self.rotamers:
            self.log.info("%s", r)

    def get_class_info(self, rot_name, class_name):
        """
        Creates a dictionary with the relevant information for the rotamer class
        :param rot_name: input rotamer class folder
        :param class_name: output rotamer class files folder
        :return: dictionary
        """
        class_dict = {'inp_dir': self.models_dir,
                      'class': class_name,
                      'class_pairs': [],
                      'reference_model': 'miss',
                      'reference_mrc': None,
                      }

        all_files = os.listdir(self.models_dir)
        all_model_names = [i for i in all_files if i.endswith('.cif')]
        all_mrc_names = [i for i in all_files if i.endswith('.mrc')]
        self.map_suffix = '_' + all_mrc_names[0].split('_')[-1]

        rotamer_dir = os.path.join(self.rotamer_classes_dir, rot_name)
        files = os.listdir(rotamer_dir)

        rotamer_names = [i for i in files if i.endswith('.cif') and not i.startswith('class')]
        for model in rotamer_names:
            if model.endswith('representative.cif'):
                tmp = model.split('_representative')[0] + '.cif'
                class_dict['reference_model'] = os.path.join(self.models_dir, tmp)
                class_dict['reference_mrc'] = os.path.join(self.models_dir, tmp.split('_')[0] + self.map_suffix)
                pair = (tmp.split('_')[0] + self.map_suffix, tmp)
                class_dict['class_pairs'].append(pair)

            if model in all_model_names:
                base_name = model.split('_')[0]
                mrc = base_name + self.map_suffix
                if mrc in all_mrc_names:
                    pair = (mrc, model)
                    class_dict['class_pairs'].append(pair)
        nr = len(class_dict['class_pairs'])
        self.log.info("\nProcessing %s\n%s models were found for the %s rotamer", rot_name, nr, rot_name)
        self.log.info("The output files will be saved in the %s folder", class_name)
        if len(rotamer_names) > 0:
            residue_type = rotamer_names[0].split('-')[0]
            motif_stats = {'residue': residue_type,
                           'rotamer': rot_name.split('_')[-1],
                           'fragments_per_entry': dict(self.count_entries(rotamer_names))
                           }
            self.statistics.append(motif_stats)

        return class_dict

    @staticmethod
    def count_entries(files_list):
        """
        Counts the number of maps fragments coming from each entry
        :param files_list: list of map or residue files
        :return: counter object
        """
        entries = []
        for file in files_list:
            entry = file.split('_')[0].split('-')[-1]
            entries.append('EMD-' + entry)
        return Counter(entries)

    def create_reference_grid(self, map_example, out_map_path):
        """
        Creates a reference map grid. All densities will be resampled on to this grid after superimposing
        :param map_example: any map from the the list of maps to be averaged
        :param out_map_path: in_dir to the grid
        """
        map_obj = MapParser(map_example)
        c = self.final_box_size
        size = int(c / map_obj.voxel_size[0])
        grid = np.zeros((size, size, size), dtype='float32')
        out_map_obj = MapParser()
        out_map_obj.data = grid
        out_map_obj.cell = (c, c, c)
        out_map_obj.origin = (-c / 2, 0, 0)
        out_map_obj.write_map(out_map_path)

    def superimpose(self, class_dict, fit_map=False):
        """
        Performs map superimposing in Chimera.
        :param class_dict: rotamer class dictionary
        :param fit_map: (Boolean) To do map to map fitting or not
        """
        class_dict['fit_map'] = fit_map
        class_dict['global_reference'] = self.get_global_reference_path()
        out_dir = class_dict['out_dir']
        inp_dir = class_dict['inp_dir']
        # Create a standard grid
        map_sample = os.path.join(inp_dir, class_dict['reference_mrc'])
        grid_file = 'grid_cub_' + str(self.final_box_size) + '.mrc'
        grid_path = os.path.join(out_dir, grid_file)
        self.create_reference_grid(map_sample, grid_path)
        class_dict['grid_path'] = grid_path

        json_path = os.path.join(out_dir, 'class_dict.json')
        with open(json_path, 'w') as j_file:
            json.dump(class_dict, j_file, indent=4)

        copy2(self.superimpose_chimera_script, out_dir)
        py_path = os.path.join(out_dir, os.path.basename(self.superimpose_chimera_script))
        command = self.chimera_path + ' --nogui ' + py_path + ' > chimera.out'
        subprocess.call(command, cwd=out_dir, shell=True)

    @staticmethod
    def normalize_truncate_neg(xyz, new_max_i):
        """
        Computes image normalisation
        Negative values are set to 0
        :param xyz: numpy array
        :param new_max_i: maximal output voxel intensity
        :return: numpy array
        """
        max_i = float(np.amax(xyz))
        xyz = xyz.clip(0, max_i) * new_max_i / max_i
        return xyz

    @staticmethod
    def normalize(xyz, new_max_i):
        """
        Computes image normalisation
        The negative and positive ranges are computed separately
        The output negative value range is dependant on the positive range
        :param xyz: numpy array
        :param new_max_i: maximal output voxel intensity
        :return: numpy array
        """
        xyz[np.isnan(xyz)] = 0
        min_i = float(np.amin(xyz))
        max_i = float(np.amax(xyz))
        if max_i != 0:
            new_min_i = new_max_i * min_i / max_i
        else:
            return None

        a = xyz.clip(0, max_i) * new_max_i / max_i
        b = (xyz.clip(min_i, 0) - min_i) * new_min_i / min_i + new_min_i

        return a + b

    @staticmethod
    def standardise(xyz, empty):
        """
        Computes image normalisation
        The negative and positive ranges are computed separately
        The output negative value range is dependant on the positive range
        :param xyz: numpy array
        :return: numpy array
        """
        std = np.std(xyz)
        mean = np.mean(xyz)
        xyz[np.isnan(xyz)] = 0
        standardised = (xyz - mean) / std

        return standardised

    def average(self, class_dict):
        """
        Map averaging
        :param class_dict: rotamer class dictionary
        """
        cl_dir = class_dict['out_dir']
        reference_map_path = class_dict['grid_path']

        if self.scaling == 'normalise-all':
            scaling_function = self.normalize
        elif self.scaling == 'normalise-positive':
            scaling_function = self.normalize_truncate_neg
        elif self.scaling == 'standardise':
            scaling_function = self.standardise
        else:
            self.log.info("Error: The normalization flag can be 'normalise-all', "
                          "'normalise-positive' or 'standardise' not %s", self.scaling)
            sys.exit()

        reference_map = MapParser(reference_map_path)
        tmp_grid = np.zeros(reference_map.n_xyz, dtype='float32')
        cella = reference_map.cell
        origin = reference_map.origin
        exceptions = 0
        for pair in class_dict['class_pairs']:
            mrc = pair[0]
            mrc_path = os.path.join(cl_dir, mrc)

            map_obj = MapParser(mrc_path)
            if map_obj.data is None:
                self.log.info("Could not read %s data", mrc_path)
                exceptions += 1
                continue
            map_obj.data = scaling_function(map_obj.data, 1)

            if map_obj.data is not None:
                if np.isnan(np.sum(map_obj.data)):
                    self.log.info("Map %s has NaN values ", mrc_path)
                    map_obj.data[np.isnan(map_obj.data)] = 0
                map_obj.write_map(mrc_path)
                try:
                    tmp_grid = tmp_grid + map_obj.data
                except ValueError:
                    self.log.info("The grid sampling of the \n%s %s\n map is different "
                                  "from the reference map \n%s %s",
                                  mrc_path, map_obj.data.shape, reference_map_path, tmp_grid.shape)
                    exceptions += 1
            else:
                exceptions += 1

        tmp_grid = tmp_grid / (len(class_dict['class_pairs']) - exceptions)
        averaged_map = MapParser()
        averaged_map.cell = cella
        averaged_map.data = np.float32(tmp_grid)
        averaged_map.origin = origin

        head, tail = os.path.split(cl_dir)
        prefix = class_dict['class_pairs'][0][0].split('-')[0] + '_' + class_dict['class']
        out_map = os.path.join(head, prefix + '_aver_' + tail + '.mrc')
        averaged_map.write_map(out_map)

        return out_map


def main():
    parser = argparse.ArgumentParser(description='Create a Mask around the atomic residue')
    parser.add_argument("-d", "--model_dir", dest="m_dir", required=True, help="Models and maps directory")
    parser.add_argument("-r", "--rot_dir", dest="r_dir", required=True, help="Rotamer classes directory")
    parser.add_argument("-p", "--out_rot_prefix", dest="rot_prefix", default='rot_', help="Rotamer folders prefix")
    parser.add_argument("-c", "--out__class_prefix", dest="class_prefix", default='class_',
                        help="Classes folders prefix")
    parser.add_argument("-b", "--box", dest="box", default=14, help="Output motifs map box size")
    parser.add_argument("-m", "--mot_dir", dest="mot", default=None, help="Output motifs directory")
    parser.add_argument("-w", "--warning_level", dest="warning_level", default='info', help="Warning level")
    parser.add_argument("-s", "--scaling", dest="scaling", default='standardise',
                        help="Image scaling normalization: 'normalise-all' - full intensity range, "
                             "'normalise-positive' - negative values are sett to 0 "
                             "'standardise' - standardisation (x-mean)/std")
    parser.add_argument("-l", "-log", dest="log", default=None, help="Log file path")
    parser.add_argument("-nr", "--rel_nr", dest="rel_nr", default=10, help="Minimum number of maps for a reliable motif")
    args = parser.parse_args()
    MapAveraging(args.m_dir, args.r_dir, args.rot_prefix, args.class_prefix, args.box,
                 args.mot, log_path=args.log, warning_level=args.warning_level,
                 scaling=args.scaling, min_reliable_nr=args.rel_nr)


if __name__ == '__main__':
    main()



#!/Users/andrei/PycharmProjects/for_testing/venv/bin/python

import threed_strudel
import os
import sys
from distutils.spawn import find_executable

config_path = os.path.abspath(os.path.join(os.path.dirname(threed_strudel.__file__), 'configure.py'))

CHIMERA_PATH = sys.argv[1]

out = find_executable(CHIMERA_PATH)
if out is None:
    print(f'ERROR: \n'
          f'{CHIMERA_PATH}'
          f'\n is not an executable')

else:
    with open(config_path, 'r') as f:
        lines = f.readlines()
        for i, line in enumerate(lines):
            words = line.split()
            if len(words) > 1:
                if words[0] == 'CHIMERA_PATH':
                    words[-1] = f'"{CHIMERA_PATH}"'
                    lines[i] = ' '.join(words) + '\n'
    with open(config_path, 'w') as f:
        f.writelines(lines)